from argparse import ArgumentParser

import torch

from ride import optimizers
from ride.core import Configs, RideModule

from .dummy_dataset import DummyRegressionDataLoader


def apply_standard_args(args):
    args.max_epochs = 3
    args.gpus = 0
    args.checkpoint_callback = True
    args.performance_metric = "loss"
    args.id = "automated_test"
    args.test_ensemble = 0
    args.loss = "mse_loss"
    args.accumulate_grad_batches = 1
    return args


def dummy_module_with_optimimzer(Optimizer):
    class DummyModule(RideModule, DummyRegressionDataLoader, Optimizer):
        def __init__(self, hparams):
            self.l1 = torch.nn.Linear(self.input_shape[0], self.hparams.hidden_dim)
            self.l2 = torch.nn.Linear(self.hparams.hidden_dim, self.output_shape)
            self.loss = torch.nn.functional.mse_loss

        def forward(self, x):
            x = x.view(x.size(0), -1)
            x = torch.relu(self.l1(x))
            x = torch.relu(self.l2(x))
            return x

        @staticmethod
        def configs():
            c = Configs()
            c.add(
                name="hidden_dim",
                type=int,
                default=128,
                strategy="choice",
                choices=[128, 256, 512, 1024],
                description="Number of hiden units.",
            )
            return c

    return DummyModule


def test_optimizers():
    for Optimizer in [
        optimizers.SgdOptimizer,
        optimizers.AdamWOptimizer,
        optimizers.SgdReduceLrOnPlateauOptimizer,
        optimizers.AdamWReduceLrOnPlateauOptimizer,
        optimizers.SgdCyclicLrOptimizer,
        optimizers.AdamWCyclicLrOptimizer,
        optimizers.SgdOneCycleOptimizer,
        optimizers.AdamWOneCycleOptimizer,
    ]:
        DummyModule = dummy_module_with_optimimzer(Optimizer)
        parser = DummyModule.configs().add_argparse_args(ArgumentParser())
        args, _ = parser.parse_known_args()
        args = apply_standard_args(args)
        module = DummyModule(args)

        ret = module.configure_optimizers()
        if isinstance(ret, tuple):
            [optimizer], s = ret
            scheduler = s["scheduler"] if isinstance(s, dict) else s[0]
            assert issubclass(type(optimizer), torch.optim.Optimizer)
            assert hasattr(scheduler, "step")
        elif isinstance(ret, dict):
            optimizer = ret["optimizer"]
            scheduler = ret["lr_scheduler"]
            monitor = ret["monitor"]
            assert issubclass(type(optimizer), torch.optim.Optimizer)
            assert hasattr(scheduler, "step")
            assert isinstance(monitor, str)
        else:
            assert issubclass(type(ret), torch.optim.Optimizer)

        assert ret is not None



from collections import OrderedDict
from typing import Tuple

import pytest
import torch

from ride import Configs, Main
from ride.core import RideModule

# from ride.feature_extraction import FeatureExtractable
from ride.optimizers import SgdOptimizer
from ride.utils.utils import AttributeDict

from .dummy_dataset import DummyRegressionDataLoader


class ExDummyModule(
    RideModule,
    DummyRegressionDataLoader,
    SgdOptimizer,
    # FeatureExtractable, # Not needed: FeatureVisualisable is part of core and inherits the functionality
):
    def __init__(self, hparams):
        self.l1 = torch.nn.Linear(self.input_shape[0], self.hparams.hidden_dim)
        self.seq1 = torch.nn.Sequential(
            torch.nn.Linear(self.hparams.hidden_dim, self.hparams.hidden_dim)
        )
        self.seq2 = torch.nn.Sequential(
            OrderedDict(
                [("l3", torch.nn.Linear(self.hparams.hidden_dim, self.output_shape))]
            )
        )
        self.loss = torch.nn.functional.mse_loss

    def forward(self, x):
        x = x.view(x.size(0), -1)
        x = torch.relu(self.l1(x))
        x = torch.relu(self.seq1(x))
        x = torch.relu(self.seq2(x))
        return x

    @staticmethod
    def configs():
        c = Configs()
        c.add(
            name="hidden_dim",
            type=int,
            default=128,
            strategy="choice",
            choices=[128, 256, 512, 1024],
            description="Number of hiden units.",
        )
        return c


@pytest.fixture()  # scope="module"
def main_and_args() -> Tuple[Main, AttributeDict]:
    m = Main(ExDummyModule)
    parser = m.argparse(run=False)
    args, _ = parser.parse_known_args()
    args.max_epochs = 1
    args.gpus = 0
    args.checkpoint_callback = True
    args.optimization_metric = "loss"
    args.id = "automated_test"
    args.test_ensemble = 0
    args.checkpoint_every_n_steps = 0
    args.monitor_lr = 0
    args.auto_lr_find = 0
    args.auto_scale_batch_size = 0
    args.num_workers = 1
    args.batch_size = 4

    return m, args


class TestFeatureExtraction:
    def test_bad_name(self, caplog, main_and_args: Tuple[Main, AttributeDict]):
        m, args = main_and_args
        args.test = True
        args.extract_features_after_layer = "badname"

        with pytest.raises(AssertionError) as e:
            m.main(args)

        assert e.value.args[0] == (
            "Invalid `extract_features_after_layer` (badname). "
            "Available layers are: ['l1', 'seq1', 'seq1.0', 'seq2', 'seq2.l3']"
        )

    def test_shallow(self, caplog, main_and_args: Tuple[Main, AttributeDict]):
        m, args = main_and_args
        args.test = True
        args.extract_features_after_layer = "l1"

        m.main(args)

        assert len(m.runner.trainer.model.extracted_features) > 1

    def test_deep(self, caplog, main_and_args: Tuple[Main, AttributeDict]):
        m, args = main_and_args
        args.test = True

        args.extract_features_after_layer = "seq1.0"
        m.main(args)

        assert len(m.runner.trainer.model.extracted_features) > 1

        args.extract_features_after_layer = "seq2.l3"
        m.main(args)
        assert len(m.runner.trainer.model.extracted_features) > 1



import io
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import pytorch_lightning as pl
from matplotlib.figure import Figure
from PIL import Image
from pytorch_lightning.loggers import (
    LightningLoggerBase,
    LoggerCollection,
    NeptuneLogger,
    TensorBoardLogger,
    WandbLogger,
)
from pytorch_lightning.utilities import rank_zero_only

from ride.metrics import FigureDict
from ride.utils.env import RUN_LOGS_PATH
from ride.utils.logging import getLogger, process_rank

logger = getLogger(__name__)
ExperimentLogger = Union[TensorBoardLogger, LoggerCollection, WandbLogger]
ExperimentLoggerCreator = Callable[[str], ExperimentLogger]


def singleton_experiment_logger() -> ExperimentLoggerCreator:
    _loggers = {}

    def experiment_logger(
        name: str = None,
        logging_backend: str = "tensorboard",
        project_name: str = None,
        save_dir=RUN_LOGS_PATH,
    ) -> ExperimentLogger:
        nonlocal _loggers
        if logging_backend not in _loggers:
            if process_rank != 0:  # pragma: no cover
                _loggers[logging_backend] = pl.loggers.base.DummyLogger()
                _loggers[logging_backend].log_dir = None
                return _loggers[logging_backend]

            logging_backend = logging_backend.lower()
            if logging_backend == "tensorboard":
                _loggers[logging_backend] = TensorBoardLogger(
                    save_dir=save_dir, name=name
                )
            elif logging_backend == "wandb":
                _loggers[logging_backend] = WandbLogger(
                    save_dir=save_dir,
                    name=name,
                    project=project_name,
                )
                _loggers[logging_backend].log_dir = getattr(
                    _loggers[logging_backend].experiment._settings, "_sync_dir", None
                )
            else:
                logger.warn("No valid logger selected.")

        return _loggers[logging_backend]

    return experiment_logger


experiment_logger = singleton_experiment_logger()


def fig2img(fig):
    """Convert a Matplotlib figure to a PIL Image and return it"""
    buf = io.BytesIO()
    fig.savefig(buf)
    buf.seek(0)
    img = Image.open(buf)
    return img


def add_experiment_logger(
    prev_logger: LightningLoggerBase, new_logger: LightningLoggerBase
) -> LoggerCollection:
    # If no logger existed previously don't do anything
    if not prev_logger:
        return None

    if isinstance(prev_logger, LoggerCollection):
        return LoggerCollection([*prev_logger._logger_iterable, new_logger])

    return LoggerCollection([prev_logger, new_logger])


def get_log_dir(module: pl.LightningModule):
    loggers = (
        module.logger if hasattr(module.logger, "__getitem__") else [module.logger]
    )
    for lgr in loggers[::-1]:  # ResultLogger would be last
        if hasattr(lgr, "log_dir"):
            return lgr.log_dir


def log_figures(module: pl.LightningModule, d: FigureDict):
    assert isinstance(module, pl.LightningModule)
    module_loggers = (
        module.logger if hasattr(module.logger, "__getitem__") else [module.logger]
    )
    image_loggers = []
    for lgr in module_loggers:
        if type(lgr) == NeptuneLogger:
            # log_image(log_name, image, step=None)
            image_loggers.append(lgr.log_image)
        elif type(lgr) == TensorBoardLogger:
            # SummaryWriter.add_figure(self, tag, figure)
            image_loggers.append(lgr.experiment.add_figure)
        elif type(lgr) == WandbLogger:
            try:
                import wandb  # noqa: F401

                wandb_log = lgr.experiment.log

                def log_figure(tag, fig):
                    im = wandb.Image(fig2img(fig), caption=tag)
                    return wandb_log({tag: im})

                image_loggers.append(log_figure)
            except ImportError:
                logger.error(
                    "Before using the WandbLogger, first install WandB using `pip install wandb`"
                )

        elif type(lgr) == ResultsLogger:
            image_loggers.append(lgr.log_figure)

    if not image_loggers:
        logger.warn(
            f"Unable to log figures {d.keys()}: No compatible logger found among {module_loggers}"
        )
        return

    for k, v in d.items():
        for log in image_loggers:
            log(k, v)


class ResultsLogger(LightningLoggerBase):
    def __init__(self, prefix="test", save_to: str = None):
        super().__init__()
        self.results = {}
        self.prefix = prefix
        self.log_dir = save_to

    def _fix_name_perfix(self, s: str, replace="test/") -> str:
        if not self.prefix:
            return s

        if s.startswith(replace):
            return f"{self.prefix}/{s[5:]}"

        return f"{self.prefix}/{s}"

    @property
    def experiment(self):
        return None

    @rank_zero_only
    def log_hyperparams(self, params):
        ...

    @rank_zero_only
    def log_metrics(self, metrics: Dict, step):
        self.results = {self._fix_name_perfix(k): float(v) for k, v in metrics.items()}

    def log_figure(self, tag: str, fig: Figure):
        if self.log_dir:
            fig_path = Path(self.log_dir) / "figures" / f"{tag}.png"
            logger.info(f"💾 Saving figure {tag} to {str(fig_path)}")
            fig_path.parent.mkdir(parents=True, exist_ok=True)
            fig.savefig(str(fig_path), bbox_inches="tight")

    @rank_zero_only
    def finalize(self, status):
        pass

    @property
    def save_dir(self) -> Optional[str]:
        return self.log_dir

    @property
    def name(self):
        return "ResultsLogger"

    @property
    def version(self):
        return "1"


StepOutputs = List[Dict[str, Any]]



from operator import attrgetter
from pathlib import Path
from typing import List

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import torch

from ride.core import Configs, RideClassificationDataset
from ride.feature_extraction import FeatureExtractable
from ride.logging import get_log_dir
from ride.metrics import FigureDict
from ride.utils.io import bump_version
from ride.utils.logging import getLogger

logger = getLogger(__name__)


class FeatureVisualisable(FeatureExtractable):
    """Adds feature visualisation capabilities to model"""

    hparams: ...

    @staticmethod
    def configs() -> Configs:
        c = FeatureExtractable.configs()
        c.add(
            name="visualise_features",
            default="",
            type=str,
            choices=["", "umap", "tsne", "pca"],
            description=(
                "Visualise extracted features using selected dimensionality reduction method. "
                "Visualisations are created only during evaluation."
            ),
        )
        return c

    def validate_attributes(self):
        for hparam in FeatureVisualisable.configs().names:
            attrgetter(f"hparams.{hparam}")(self)

    def __init__(self, hparams, *args, **kwargs):
        self.dimensionality_reduction = None
        if self.hparams.visualise_features == "umap":
            try:
                from umap import UMAP

                self.dimensionality_reduction = UMAP(n_components=2)
            except ModuleNotFoundError as e:  # pragma: no cover
                logger.error(
                    "To visualise features with UMAP, first install Umap via `pip install umap-learn` or `pip install 'ride[extras]'`"
                )
                raise e

        elif self.hparams.visualise_features == "tsne":
            try:
                from sklearn.manifold import TSNE

                self.dimensionality_reduction = TSNE(n_components=2)
            except ModuleNotFoundError as e:  # pragma: no cover
                logger.error(
                    "To visualise features with TSNE, first install Scikit-learn via `pip install scikit-learn` or `pip install 'ride[extras]'`"
                )
                raise e
        elif self.hparams.visualise_features == "pca":
            try:
                from sklearn.decomposition import PCA

                self.dimensionality_reduction = PCA(n_components=2)
            except ModuleNotFoundError as e:  # pragma: no cover
                logger.error(
                    "To visualise features with PCA, first install Scikit-learn via `pip install scikit-learn` or `pip install 'ride[extras]'`"
                )
                raise e

        if (
            self.dimensionality_reduction
            and not self.hparams.extract_features_after_layer
        ):
            logger.error(
                "Unable to visualise features if no layer is specified using `extract_features_after_layer`."
            )

    def metrics_epoch(
        self,
        preds: torch.Tensor,
        targets: torch.Tensor,
        prefix: str = None,
        *args,
        **kwargs,
    ) -> FigureDict:
        if not hasattr(self, "extracted_features"):
            return {}

        FeatureExtractable.metrics_epoch(
            self, preds, targets, prefix, clear_extracted_features=(prefix != "test")
        )
        if (
            prefix != "test"
            or not self.dimensionality_reduction
            or not len(self.extracted_features) > 0
        ):
            return {}

        # Dimensionality reduction
        try:
            feat = np.stack(self.extracted_features)
            if len(feat.shape) > 2:
                logger.debug(
                    f"🔧 Flattening extracted_features ({feat.shape[0]} -> {np.prod(feat.shape[1:])}) prior to dimensionality reduction."
                )
                feat = feat.reshape(feat.shape[0], -1)

            logger.info(
                f"👁 Performing dimensionality reduction using {self.hparams.visualise_features.upper()}"
            )
            features = self.dimensionality_reduction.fit_transform(feat)

            # Save features
            base_path = Path(get_log_dir(self)) / "features" / (prefix or "")
            base_path.mkdir(parents=True, exist_ok=True)
            base_name = self.hparams.extract_features_after_layer.replace(".", "_")
            save_path = bump_version(
                base_path / f"{base_name}_{self.hparams.visualise_features}.npy"
            )
            logger.info(
                f"💾 Saving {self.hparams.visualise_features.upper()} features to {str(save_path)}"
            )
            np.save(save_path, features)

            # Create scatterplot
            fig = (
                scatter_plot(features, np.array(targets), self.classes)
                if issubclass(type(self), RideClassificationDataset)
                else scatter_plot(features)
            )
            return {
                f"{self.hparams.extract_features_after_layer}_{self.hparams.visualise_features}": fig
            }

        except Exception as e:
            logger.error(f"Caught exception during feature visualisation: {e}")
            return {}


def scatter_plot(
    features: np.array, labels: np.array = None, classes: List[str] = None
):
    sns.set_theme()

    fig = plt.figure(figsize=(6, 6))

    if labels is not None:
        palette = sns.color_palette("hls", n_colors=len(classes))
        legend = True
        plt.legend(loc="center left", bbox_to_anchor=(0.997, 0.5))
    else:
        palette = None
        legend = False

    # Marker size according to number of features (heuristic choice)
    s = max(3, round((20 - len(features) / 150)))

    g = sns.scatterplot(
        x=features[:, 0],
        y=features[:, 1],
        hue=labels,
        linewidth=0,
        alpha=0.7,
        palette=palette,
        s=s,
        legend=legend,
    )

    if legend:
        fig.axes[0].legend(loc="center left", bbox_to_anchor=(0.997, 0.5))
        for t, l in zip(g.legend_.texts, classes):
            t.set_text(l)
    g.set(xticklabels=[])
    g.set(yticklabels=[])
    plt.axis("equal")
    plt.tight_layout()

    return fig



# from ride.profile import Profileable
import inspect
from abc import ABC
from typing import Any, List, Sequence, Union

import pytorch_lightning as pl
from corider import Configs as _Configs
from pytorch_lightning.utilities.parsing import AttributeDict
from supers import supers
from torch import Tensor
from torch.utils.data import DataLoader

from ride.utils.logging import getLogger
from ride.utils.utils import (
    DictLike,
    attributedict,
    is_shape,
    merge_attributedicts,
    missing_or_not_in_other,
    name,
    some,
)

logger = getLogger(__name__)


DataShape = Union[int, Sequence[int], Sequence[Sequence[int]]]


class Configs(_Configs):
    """Configs module for holding project configurations.

    This is a wrapper of the Configs found as a stand-alone package in https://github.com/LukasHedegaard/co-rider
    """

    @staticmethod
    def collect(cls: "RideModule") -> "Configs":
        """Collect the configs from all class bases

        Returns:
            Configs: Aggregated configurations
        """
        c: Configs = sum([c.configs() for c in cls.__bases__ if hasattr(c, "configs")])  # type: ignore
        return c

    def default_values(self):
        return attributedict({k: v.default for k, v in self.values.items()})


def _init_subclass(cls):
    # Validate inheritance order
    assert (
        cls.__bases__[0] == RideModule or cls.__bases__[0].__bases__[0] == RideModule
    ), """RideModule must come first in inheritance order, e.g.:
    class YourModule(RideModule, OtherMixin):
        ..."""

    add_bases = []

    # Extend funtionality with additional base-classes
    from ride.feature_visualisation import FeatureVisualisable
    from ride.finetune import Finetunable
    from ride.lifecycle import Lifecycle  # Break cyclical dependencies

    # Ensure pl.LightningModule is the lowest-priority parent
    if not cls.__bases__[-1] == pl.LightningModule:
        add_bases.append(pl.LightningModule)

    if not issubclass(cls, DefaultMethods):
        add_bases.append(DefaultMethods)

    if not issubclass(cls, Lifecycle):
        add_bases.append(Lifecycle)

    if not issubclass(cls, Finetunable):
        add_bases.append(Finetunable)

    if not issubclass(cls, FeatureVisualisable):
        add_bases.append(FeatureVisualisable)

    # Warn if there is no forward
    if missing_or_not_in_other(
        cls, pl.LightningModule, {"forward"}, must_be_callable=True
    ):
        logger.warning(
            f"No `forward` function found in {name(cls)}. Did you forget to define it?"
        )

    # Ensure dataset
    dataset_steps = {"train_dataloader", "val_dataloader", "test_dataloader"}
    missing_dataset_steps = missing_or_not_in_other(
        cls, pl.LightningModule, dataset_steps
    )
    if missing_dataset_steps:
        logger.warning(
            f"No dataloader funcions {missing_dataset_steps} found in {name(cls)}"
        )
        logger.info(
            "🔧 Adding ride.RideDataset automatically and assuming that `self.datamodule`, `self.input_shape`, and `self.output_shape` will be provided by user"
        )
        add_bases.append(RideDataset)

    # Ensure optimizer
    if missing_or_not_in_other(cls, pl.LightningModule, {"configure_optimizers"}):
        logger.info(f"`configure_optimizers` not found in in {name(cls)}")
        logger.info("🔧 Adding ride.SgdOptimizer automatically")

        from ride.optimizers import SgdOptimizer  # Avoid cyclical import

        add_bases.append(SgdOptimizer)

    # Update class bases with pl.LightningModule as lowest rank
    cls.__bases__ = (*cls.__bases__, *add_bases[::-1])

    # Monkeypatch derived module init
    cls._orig_init = cls.__init__

    def init(self, hparams: DictLike = {}, *args, **kwargs):
        pl.LightningModule.__init__(self)
        self.hparams = merge_attributedicts(self.configs().default_values(), hparams)
        supers(self)[1:-1].__init__(self.hparams)
        apply_init_args(cls._orig_init, self, self.hparams, *args, **kwargs)
        supers(self).on_init_end(self.hparams, *args, **kwargs)
        supers(self).validate_attributes()

    cls.__init__ = init

    # Monkeypatch derived module configs
    orig_configs = getattr(cls, "configs", None)

    @staticmethod
    def configs():
        c = Configs.collect(cls)
        if orig_configs:
            c += orig_configs()
        return c

    cls.configs = configs


def apply_init_args(fn, self, hparams, *args, **kwargs):
    spec = inspect.getfullargspec(fn)
    valid_kwargs = (
        kwargs
        if spec.varkw == "kwargs"
        else {k: v for k, v in kwargs.items() if k in spec.args}
    )
    if len(spec.args) == 1:
        return fn(self)
    return fn(self, hparams, *args, **valid_kwargs)


class RideModule:
    """
    Base-class for modules using the Ride ecosystem.

    This module should be inherited as the highest-priority parent (first in sequence).

    Example::

        class MyModule(ride.RideModule, ride.SgdOneCycleOptimizer):
            def __init__(self, hparams):
                ...

    It handles proper initialisation of `RideMixin` parents and adds automatic attribute validation.

    If `pytorch_lightning.LightningModule` is omitted as lowest-priority parent, `RideModule` will automatically add it.

    If `training_step`, `validation_step`, and `test_step` methods are not found, the `ride.Lifecycle` will be automatically mixed in by this module.
    """

    def __init_subclass__(cls):
        # Only initialise immediate children
        if cls.__bases__[0] == RideModule:
            _init_subclass(cls)

    @property
    def hparams(self) -> AttributeDict:
        if not hasattr(self, "_hparams"):
            self._hparams = AttributeDict()
        return self._hparams

    @hparams.setter
    def hparams(self, hp: Union[dict, AttributeDict, Any]):
        # Overload the version in pytorch_lightning core to omit DeprecationWarning
        self._hparams = attributedict(hp)

    @classmethod
    def with_dataset(cls, ds: "RideDataset"):
        new_bases = [b for b in cls.__bases__ if not issubclass(b, RideDataset)]
        old_dataset = [b for b in cls.__bases__ if issubclass(b, RideDataset)]
        assert len(old_dataset) <= 1, "`RideModule` should only have one `RideDataset`"
        if old_dataset and issubclass(old_dataset[0], RideClassificationDataset):
            assert issubclass(
                ds, RideClassificationDataset
            ), "A `RideClassificationDataset` should be replaced by a `RideClassificationDataset`"
        new_bases.insert(-1, ds)
        cls.__init__ = cls._orig_init  # Revert to orig init
        DerivedRideModule = type(
            f"{name(cls)}With{name(ds)}", tuple(new_bases), dict(cls.__dict__)
        )

        return DerivedRideModule


class RideMixin(ABC):
    """Abstract base-class for Ride mixins"""

    def __init__(self, hparams: AttributeDict, *args, **kwargs):
        ...

    def on_init_end(self, hparams: AttributeDict, *args, **kwargs):
        ...

    def validate_attributes(self):
        ...


class DefaultMethods(RideMixin):
    def warm_up(self, input_shape: Sequence[int], *args, **kwargs):
        """Warms up the model state with a dummy input of shape `input_shape`.
        This method is called prior to model profiling.

        Args:
            input_shape (Sequence[int]): input shape with which to warm the model up, including batch size.
        """
        ...


class OptimizerMixin(RideMixin):
    """Abstract base-class for Optimizer mixins"""

    ...


class RideDataset(RideMixin):
    """Base-class for Ride datasets.

    If no dataset is specified otherwise, this mixin is automatically add as a base of RideModule childen.

    User-specified datasets must inherit from this class, and specify the following:
    - `self.input_shape`: Union[int, Sequence[int], Sequence[Sequence[int]]]
    - `self.output_shape`: Union[int, Sequence[int], Sequence[Sequence[int]]]

    and either the functions:
    - `train_dataloader`: Callable[[Any], DataLoader]
    - `val_dataloader`: Callable[[Any], DataLoader]
    - `test_dataloader`: Callable[[Any], DataLoader]

    or:
    - `self.datamodule`, which has `train_dataloader`, `val_dataloader`, and `test_dataloader` attributes.
    """

    input_shape: DataShape
    output_shape: DataShape

    def validate_attributes(self):
        assert is_shape(
            getattr(self, "input_shape", None)
        ), "RideDataset should define an `input_shape` of type int, list, tuple, or namedtuple."
        assert is_shape(
            getattr(self, "output_shape", None)
        ), "RideDataset should define `output_shape` of type int, list, tuple, or namedtuple."

        for n in RideDataset.configs().names:
            assert some(
                self, f"hparams.{n}"
            ), "`self.hparams.{n}` not found in Dataset. Did you forget to include its `configs`?"

    @staticmethod
    def configs() -> Configs:
        c = Configs()
        c.add(
            name="batch_size",
            type=int,
            default=16,
            strategy="constant",
            description="Batch size for dataset.",
        )
        c.add(
            name="num_workers",
            type=int,
            default=0,
            strategy="constant",
            description="Number of workers in dataloader.",
        )
        return c

    def train_dataloader(self, *args: Any, **kwargs: Any) -> DataLoader:
        """The train dataloader"""
        assert some(
            self, "datamodule.train_dataloader"
        ), f"{name(self)} should either have a `self.datamodule: pl.LightningDataModule` or overload the `train_dataloader` function."
        return self.datamodule.train_dataloader

    def val_dataloader(
        self, *args: Any, **kwargs: Any
    ) -> Union[DataLoader, List[DataLoader]]:
        """The val dataloader"""
        assert some(
            self, "datamodule.val_dataloader"
        ), f"{name(self)} should either have a `self.datamodule: pl.LightningDataModule` or overload the `val_dataloader` function."
        return self.datamodule.val_dataloader

    def test_dataloader(
        self, *args: Any, **kwargs: Any
    ) -> Union[DataLoader, List[DataLoader]]:
        """The test dataloader"""
        assert some(
            self, "datamodule.test_dataloader"
        ), f"{name(self)} should either have a `self.datamodule: pl.LightningDataModule` or overload the `test_dataloader` function."
        return self.datamodule.test_dataloader


class RideClassificationDataset(RideDataset):
    """Base-class for Ride classification datasets.

    If no dataset is specified otherwise, this mixin is automatically add as a base of RideModule childen.

    User-specified datasets must inherit from this class, and specify the following:
    - `self.input_shape`: Union[int, Sequence[int], Sequence[Sequence[int]]]
    - `self.output_shape`: Union[int, Sequence[int], Sequence[Sequence[int]]]
    - `self.classes`: List[str]

    and either the functions:
    - `train_dataloader`: Callable[[Any], DataLoader]
    - `val_dataloader`: Callable[[Any], DataLoader]
    - `test_dataloader`: Callable[[Any], DataLoader]

    or:
    - `self.datamodule`, which has `train_dataloader`, `val_dataloader`, and `test_dataloader` attributes.
    """

    classes: List[str]

    @property
    def num_classes(self) -> int:
        return len(self.classes)

    @staticmethod
    def configs() -> Configs:
        c = RideDataset.configs()
        c.add(
            name="test_confusion_matrix",
            type=int,
            default=0,
            choices=[0, 1],
            strategy="constant",
            description="Create and save confusion matrix for test data.",
        )
        return c

    def validate_attributes(self):
        RideDataset.validate_attributes(self)
        assert type(getattr(self, "classes", None)) in {
            list,
            tuple,
        }, "Ride RideClassificationDataset should define `classes` but none was found."

    def metrics_epoch(
        self,
        preds: Tensor,
        targets: Tensor,
        prefix: str = None,
        *args,
        **kwargs,
    ):  # -> "FigureDict":
        if prefix != "test" or not self.hparams.test_confusion_matrix:
            return {}
        from ride.metrics import make_confusion_matrix

        fig = make_confusion_matrix(preds, targets, self.classes)
        return {"confusion_matrix": fig}



import sys
from functools import partial
from pathlib import Path
from typing import Type, Union

from pytorch_lightning.utilities.parsing import AttributeDict

from ride.core import Configs, RideModule
from ride.runner import Runner, is_runnable
from ride.utils.env import NUM_CPU, TUNE_LOGS_PATH
from ride.utils.gpus import parse_num_gpus
from ride.utils.io import bump_version, dump_json, dump_yaml, load_structured_data
from ride.utils.logging import getLogger

logger = getLogger(__name__)


class Hparamsearch:
    def __init__(self, Module: Type[RideModule]):
        assert is_runnable(Module)
        self.Module = Module
        self.module_name = Module.__name__

    def configs(self) -> Configs:
        c = Configs()
        c.add(
            name="trials",
            default=30,
            type=int,
            description="Number of trials in the hyperparameter search",
        )
        c.add(
            name="gpus_per_trial",
            default=0,
            type=int,
            description="Number of GPUs per trail in the hyperparameter search",
        )
        c.add(
            name="optimization_metric",
            default="loss",
            type=str,
            choices=self.Module.metric_names(),
            description="Name of the performance metric that should be optimized",
        )
        c.add(
            name="from_hparam_space_file",
            default=None,
            type=str,
            description="Path to file with specification for the search space during hyper-parameter optimisation",
        )
        return c

    def __call__(self, args: AttributeDict):
        self.run(args)

    def run(self, args: AttributeDict):
        """Run hyperparameter search using the `tune.schedulers.ASHAScheduler`

        Args:
            args (AttributeDict): Arguments

        Side-effects:
            Saves logs to `TUNE_LOGS_PATH / args.id`
        """
        try:
            from ray import tune
            from ray.tune.integration.pytorch_lightning import (
                TuneReportCheckpointCallback,
            )
        except ModuleNotFoundError as e:  # pragma: no cover
            logger.error(
                "To use hyperparameter search, first install Ray Tune via `pip install 'ray[tune]'` or `pip install 'ride[extras]'`"
            )
            raise e

        if not hasattr(args, "id"):
            args.id = "hparamsearch"

        module_config = (
            Configs.from_file(args.from_hparam_space_file)
            if args.from_hparam_space_file
            else self.Module.configs()
        ).tune_config()

        config = {
            **dict(args),
            **module_config,
            # pl.Trainer args:
            "gpus": args.gpus_per_trial,
            "logger": False,
            "accumulate_grad_batches": (
                (8 // args.gpus_per_trial) * args.accumulate_grad_batches
                if args.gpus_per_trial
                else args.accumulate_grad_batches
            ),
        }
        scheduler = tune.schedulers.ASHAScheduler(
            metric=f"val/{args.optimization_metric}",
            mode=self.Module.metrics()[args.optimization_metric].value,
            max_t=args.max_epochs,
            grace_period=1,
            reduction_factor=2,
        )

        metric_names = [f"val/{m}" for m in self.Module.metrics().keys()]

        reporter = tune.CLIReporter(
            metric_columns=[*metric_names, "training_iteration"],
        )
        tune_callbacks = [
            TuneReportCheckpointCallback(
                metrics=metric_names,
                filename="checkpoint",
                on="validation_end",
            )
        ]
        cpus_per_trial = max(
            1,
            (
                min(10 * args.gpus_per_trial, NUM_CPU - 10)
                if args.gpus_per_trial
                else min(10, NUM_CPU - 2)
            ),
        )

        analysis = tune.run(
            partial(
                Runner.static_train_and_val,
                self.Module,
                trainer_callbacks=tune_callbacks,
            ),
            name=args.id,
            local_dir=str(TUNE_LOGS_PATH),
            resources_per_trial={"cpu": cpus_per_trial, "gpu": args.gpus_per_trial},
            config=config,
            num_samples=args.trials,
            scheduler=scheduler,
            progress_reporter=reporter,
            queue_trials=False,
            raise_on_failed_trial=False,
        )

        best_hparams = analysis.get_best_config(
            metric=f"val/{args.optimization_metric}",
            mode=self.Module.metrics()[args.optimization_metric].value,
            scope="all",
        )
        # Select only model parameters
        if best_hparams:
            best_hparams = {
                k: best_hparams[k]
                for k in [
                    *self.Module.configs().names,
                    # Trainer parameters that influence model hparams:
                    "accumulate_grad_batches",
                    "batch_size",
                    "gpus",
                ]
            }
        return best_hparams

    @staticmethod
    def dump(hparams: dict, identifier: str, extention="yaml") -> str:
        """Dumps haparams to TUNE_LOGS_PATH / identifier / "best_hparams.json" """
        dump_path = bump_version(
            TUNE_LOGS_PATH / identifier / f"best_hparams.{extention}"
        )
        dump = {"json": dump_json, "yaml": dump_yaml}[extention]
        dump(dump_path, hparams)
        return str(dump_path)

    @staticmethod
    def load(
        path: Union[Path, str],
        old_args=AttributeDict(),
        Cls: Type[RideModule] = None,
        auto_scale_lr=False,
    ) -> AttributeDict:
        """Loads hparams from path

        Args:
            path (Union[Path, str]): Path to jsonfile containing hparams
            old_args (Optional[AttributeDict]):The AttributeDict to be updated with the new hparams
            cls (Optional[RideModule]): A class whole hyperparameters can be used to select the relevant hparams to take

        Returns:
            AttributeDict: AttributeDict with updated hyperparameters
        """
        path = Path(path)
        hparams = load_structured_data(path)

        if Cls:
            hparam_names = Cls.configs().names
            hparams = {k: v for k, v in hparams.items() if k in hparam_names}

        # During hparamsearch, only a single GPU is used, but accumulate_grad_batches is set to the total number of gpus given
        # If we have multiple GPUs, we need to reduce accumulate_grad_batches accordingly
        num_gpus = parse_num_gpus(old_args.gpus)
        if num_gpus > 0 and "accumulate_grad_batches" in hparams:  # pragma: no cover
            hparams["accumulate_grad_batches"] = max(
                1, int(hparams["accumulate_grad_batches"]) // num_gpus
            )

        old_args = dict(old_args)
        user_passed_arg_keys = [a[2:] for a in sys.argv if a.startswith("--")]
        user_passed_args = {
            k: v for k, v in old_args.items() if k in user_passed_arg_keys
        }

        # If batch size was changed by user, automatically apply the linear scaling rule to the learning rate
        if (
            auto_scale_lr
            and "batch_size" in hparams
            and "learning_rate" in hparams
            and "batch_size" in user_passed_args
            and "learning_rate" not in user_passed_args
        ):
            old_accumulate_grad_batches = (
                hparams["accumulate_grad_batches"]
                if "accumulate_grad_batches" in hparams
                else 1
            )
            new_accumulate_grad_batches = (
                user_passed_args["accumulate_grad_batches"]
                if "accumulate_grad_batches" in user_passed_args
                else old_accumulate_grad_batches
            )
            new_tot_batch = new_accumulate_grad_batches * user_passed_args["batch_size"]
            old_tot_batch = old_accumulate_grad_batches * hparams["batch_size"]
            if new_tot_batch != old_tot_batch:
                scaling = new_tot_batch / old_tot_batch
                user_passed_args["learning_rate"] = hparams["learning_rate"] * scaling
                logger.info(
                    f"🔧 A `batch_size*accumulate_grad_batches` ({new_tot_batch}) differs from hparams file ({old_tot_batch}). "
                    f"Scaling learning_rate from {hparams['learning_rate']} to {user_passed_args['learning_rate']} (= {hparams['learning_rate']} × {new_tot_batch} / {old_tot_batch})"
                )

        return AttributeDict(**{**old_args, **hparams, **user_passed_args})



from operator import attrgetter
from pathlib import Path

import numpy
import torch

from ride.core import Configs, RideMixin
from ride.logging import get_log_dir
from ride.metrics import MetricDict
from ride.utils.io import bump_version
from ride.utils.logging import getLogger
from ride.utils.utils import rgetattr

logger = getLogger(__name__)


class FeatureExtractable(RideMixin):
    """Adds feature extraction capabilities to model"""

    hparams: ...

    @staticmethod
    def configs() -> Configs:
        c = Configs()
        c.add(
            name="extract_features_after_layer",
            default="",
            type=str,
            description=(
                "Layer name after which to extract features. "
                "Nested layers may be selected using dot-notation, "
                "e.g. `block.subblock.layer1`"
            ),
        )
        return c

    def validate_attributes(self):
        for hparam in FeatureExtractable.configs().names:
            attrgetter(f"hparams.{hparam}")(self)

    def on_init_end(self, hparams, *args, **kwargs):
        if not self.hparams.extract_features_after_layer:
            return

        available_layers = [k for k, _ in self.named_modules() if k != ""]

        assert self.hparams.extract_features_after_layer in available_layers, (
            f"Invalid `extract_features_after_layer` ({self.hparams.extract_features_after_layer}). "
            f"Available layers are: {available_layers}"
        )

        layer = rgetattr(self, self.hparams.extract_features_after_layer)

        self.extracted_features = []

        def store_features(sself, input, output):
            nonlocal self
            for o in output:
                self.extracted_features.append(o.detach().cpu().numpy())

        layer.register_forward_hook(store_features)

    def metrics_epoch(
        self,
        preds: torch.Tensor,
        targets: torch.Tensor,
        prefix: str = None,
        clear_extracted_features=True,
        *args,
        **kwargs,
    ) -> MetricDict:
        if not hasattr(self, "extracted_features"):
            return {}

        # Save
        save_path = bump_version(
            Path(get_log_dir(self))
            / "features"
            / (prefix or "")
            / f"{self.hparams.extract_features_after_layer.replace('.','_')}.npy"
        )
        save_path.parent.mkdir(parents=True, exist_ok=True)
        logger.info(f"💾 Saving extracted features to {str(save_path)}")
        numpy.save(save_path, self.extracted_features)

        if clear_extracted_features and prefix != "test":
            self.extracted_features = []

        return {}



import numpy as np
import torch

import ride
from examples.mnist_dataset import MnistDataset


class SimpleClassifier(
    ride.RideModule,
    ride.SgdOneCycleOptimizer,
    ride.TopKAccuracyMetric(1, 3),
    MnistDataset,
):
    def __init__(self, hparams):
        # `self.input_shape` and `self.output_shape` were injected via `MnistDataset`
        self.l1 = torch.nn.Linear(np.prod(self.input_shape), self.hparams.hidden_dim)
        self.l2 = torch.nn.Linear(self.hparams.hidden_dim, self.output_shape)

    def forward(self, x):
        x = x.view(x.size(0), -1)
        x = torch.relu(self.l1(x))
        x = torch.relu(self.l2(x))
        return x

    @staticmethod
    def configs():
        c = ride.Configs()
        c.add(
            name="hidden_dim",
            type=int,
            default=128,
            strategy="choice",
            choices=[128, 256, 512, 1024],
            description="Number of hidden units.",
        )
        return c


if __name__ == "__main__":
    ride.Main(SimpleClassifier).argparse()



from collections import namedtuple

from ride.utils.utils import is_shape


def test_is_shape():
    # OK
    assert is_shape(42)  # int
    assert is_shape([42, 13])  # list of ints
    assert is_shape((42,))  # tuple of ints

    MyShape = namedtuple("MyShape", ["x", "y"])
    assert is_shape(MyShape(42, 13))  # namedtuple of ints

    # Not OK
    class MyClass:
        ...

    assert not is_shape(MyClass)  # regular class
    assert not is_shape(42.0)  # float
    assert not is_shape("42")  # str
    assert not is_shape((42, "13"))  # tuple of mixed types



# Modified from https://github.com/vdouet/Discriminative-learning-rates-PyTorch

import functools
from typing import Union

import numpy as np
import torch.nn as nn

from ride.utils.logging import getLogger

logger = getLogger(__name__)
"""
Developped by the Fastai team for the Fastai library
From the fastai library
https://www.fast.ai and https://github.com/fastai/fastai
"""

###############################################################################
# Unmodified classes and functions:


class PrePostInitMeta(type):
    "A metaclass that calls optional `__pre_init__` and `__post_init__` methods"

    def __new__(cls, name, bases, dct):
        x = super().__new__(cls, name, bases, dct)
        old_init = x.__init__

        def _pass(self):
            pass

        @functools.wraps(old_init)
        def _init(self, *args, **kwargs):
            self.__pre_init__()
            old_init(self, *args, **kwargs)
            self.__post_init__()

        x.__init__ = _init
        if not hasattr(x, "__pre_init__"):
            x.__pre_init__ = _pass
        if not hasattr(x, "__post_init__"):
            x.__post_init__ = _pass
        return x


class Module(nn.Module, metaclass=PrePostInitMeta):
    "Same as `nn.Module`, but no need for subclasses to call `super().__init__`"

    def __pre_init__(self):
        super().__init__()

    def __init__(self):
        pass


class ParameterModule(Module):
    "Register a lone parameter `p` in a module."

    def __init__(self, p: nn.Parameter):
        self.val = p

    def forward(self, x):
        return x


def children(m: nn.Module):
    "Get children of `m`."
    return list(m.children())


def num_children(m: nn.Module):
    "Get number of children modules in `m`."
    return len(children(m))


def children_and_parameters(m: nn.Module):
    "Return the children of `m` and its direct parameters not registered in modules."
    children = list(m.children())
    children_p = sum([[id(p) for p in c.parameters()] for c in m.children()], [])
    for p in m.parameters():
        if id(p) not in children_p:
            children.append(ParameterModule(p))
    return children


def even_mults(start: float, stop: float, n: int) -> np.ndarray:
    "Build log-stepped array from `start` to `stop` in `n` steps."
    mult = stop / start
    step = mult ** (1 / (n - 1))
    return np.array([start * (step ** i) for i in range(n)])


flatten_model = (
    lambda m: sum(map(flatten_model, children_and_parameters(m)), [])
    if num_children(m)
    else [m]
)
###############################################################################

"""
Modified version of lr_range from fastai
https://github.com/fastai/fastai/blob/master/fastai/basic_train.py#L185
"""


def lr_range(net: nn.Module, lr: slice, model_len: int) -> np.ndarray:
    "Build differential learning rates from `lr`."

    if not isinstance(lr, slice):
        return lr
    if lr.start:
        res = even_mults(lr.start, lr.stop, model_len)
    else:
        res = [lr.stop / 10] * (model_len - 1) + [lr.stop]
    return res


def unfreeze_layers(model: nn.Sequential, unfreeze: bool = True) -> None:
    "Unfreeze or freeze all layers"

    for layer in model.parameters():
        layer.requires_grad = unfreeze


def build_param_dicts(
    layers: nn.Sequential, lr: list = [0], return_len: bool = False
) -> Union[int, list]:
    """
    Either return the number of layers with requires_grad is True
    or return a list of dictionnaries containing each layers on its associated LR"
    Both weight and bias are check for requires_grad is True
    """

    params = []
    idx = 0
    for layer in layers:
        param = []
        if hasattr(layer, "requires_grad"):  # and layer.requires_grad:
            # To implement for custom nn.Parameter()
            logger.debug("Custom nn.Parameter() not supported")
        if getattr(layer, "weight", None) is not None:
            param.append(layer.weight)
        if getattr(layer, "bias", None) is not None:
            param.append(layer.bias)
        if param:
            params.append({"params": param, "lr": f"{lr[idx]}"})
            idx += 1
        if return_len:
            idx = 0  # We don't want to increment idx here.

    return len(params) if return_len else params


def discriminative_lr(
    net: nn.Module, lr: slice, unfreeze: bool = False
) -> Union[list, np.ndarray, nn.Sequential]:
    """
    Flatten our model and generate a list of dictionnaries to be passed to the
    optimizer.
    - If only one learning rate is passed as a slice the last layer will have the
    corresponding learning rate and all other ones will have lr/10
    - If two learning rates are passed such as slice(min_lr, max_lr) the last
    layer will have max_lr as a learning rate and the first one will have min_lr.
    All middle layers will have learning rates logarithmically interpolated
    ranging from min_lr to max_lr
    """

    layers = nn.Sequential(*flatten_model(net))  # Flatten/ungroup our model
    if unfreeze:
        unfreeze_layers(layers, True)  # Unfreeze all layers

    # Return the number of layer where requires_grad is True (bias + weight)
    model_len = build_param_dicts(layers, return_len=True)

    # Create the list of learning rates
    list_lr = lr_range(net, lr, model_len)

    # Create our optimizer parameters list of dictionnaries
    params_layers = build_param_dicts(layers, list_lr)

    return params_layers, list(list_lr), layers



from collections import abc
from enum import Enum
from operator import attrgetter
from typing import Dict, Iterable, List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import torch
from matplotlib.figure import Figure
from ptflops import get_model_complexity_info
from supers import supers
from torch import Tensor
from torchmetrics.functional.classification import average_precision
from torchmetrics.functional.classification.confusion_matrix import confusion_matrix

from ride.core import Configs, RideMixin
from ride.utils.logging import getLogger
from ride.utils.utils import merge_dicts, name

ExtendedMetricDict = Dict[str, Union[Tensor, Figure]]
MetricDict = Dict[str, Tensor]
FigureDict = Dict[str, Figure]
StepOutputs = List[Dict[str, Tensor]]

logger = getLogger(__name__, log_once=True)


def sort_out_figures(d: ExtendedMetricDict) -> Tuple[MetricDict, FigureDict]:
    mets, figs = {}, {}
    for k, v in d.items():
        if type(v) == Figure:
            figs[k] = v
        else:
            mets[k] = v
    return mets, figs


class OptimisationDirection(Enum):
    MIN = "min"
    MAX = "max"


class MetricMixin(RideMixin):
    """Abstract base class for Ride modules"""

    def __init_subclass__(cls):
        if not hasattr(cls, "_metrics"):
            logger.error(
                f"Subclasses of MetricMixin should define a `_metrics` classmethod, but none was found in {name(cls)}"
            )

    @classmethod
    def metrics(cls) -> Dict[str, str]:
        ms = merge_dicts(
            *[c._metrics() for c in cls.__bases__ if issubclass(c, MetricMixin)]
        )
        return ms

    @classmethod
    def metric_names(cls) -> List[str]:
        return list(sorted(cls.metrics().keys()))

    def metrics_step(self, *args, **kwargs) -> MetricDict:
        return {}  # pragma: no cover

    def metrics_epoch(
        self, preds: Tensor, targets: Tensor, prefix: str = "", *args, **kwargs
    ) -> MetricDict:
        return {}

    def collect_metrics(self, preds: Tensor, targets: Tensor) -> MetricDict:
        mdlist: List[MetricDict] = supers(self).metrics_step(preds, targets)  # type: ignore
        return {
            k: v.to(device=self.device) if hasattr(v, "to") else v
            for md in mdlist
            for k, v in md.items()
        }

    def collect_epoch_metrics(
        self, preds: Tensor, targets: Tensor, prefix: str = None
    ) -> ExtendedMetricDict:
        mdlist: List[ExtendedMetricDict] = supers(self).metrics_epoch(preds, targets, prefix=prefix)  # type: ignore
        return {
            k: v.to(device=self.device) if hasattr(v, "to") else v
            for md in mdlist
            for k, v in md.items()
        }


def MetricSelector(  # noqa: C901
    mapping: Dict[str, Union[MetricMixin, Iterable[MetricMixin]]] = None,
    **kwargs: Union[MetricMixin, Iterable[MetricMixin]],
) -> MetricMixin:
    if not isinstance(mapping, dict):
        mapping = {}

    mapping = {**mapping, **kwargs}
    # Ensure mapping is Dict[str, List[MetricMixin]]
    mapping = {
        k: (list(v) if isinstance(v, abc.Iterable) else [v]) for k, v in mapping.items()
    }
    metric_set = set([item for sublist in mapping.values() for item in sublist])
    assert all(
        issubclass(M, MetricMixin) for M in metric_set
    ), "All passed values should be of type ride.metrics.MetricMixin"

    class MetricSelectorMixin(MetricMixin):
        @staticmethod
        def configs() -> Configs:
            c = Configs()
            c.add(
                name="metric_selection",
                default="",
                type=str,
                strategy="constant",
                description="Selection key for MetricSelector.",
                choices=list(mapping.keys()),
            )
            for Metric in metric_set:
                if hasattr(Metric, "configs"):
                    c += Metric.configs()
            return c

        @classmethod
        def _metrics(cls):
            ms = {}
            for Metric in metric_set:
                ms = {**ms, **Metric._metrics()}
            return ms

        def __init__(self, hparams, *args, **kwargs):
            assert (
                self.hparams.metric_selection in mapping
            ), f"You must specify a `metric_selection` hyperparameter. Choices: {list(mapping.keys())}"
            self.metrics_selection = mapping[self.hparams.metric_selection]
            for m in self.metrics_selection:
                m.__init__(self, hparams, *args, **kwargs)

        def on_init_end(self, *args, **kwargs):
            for m in self.metrics_selection:
                m.on_init_end(self, *args, **kwargs)

        def metrics_step(self, preds: Tensor, targets: Tensor, **kwargs) -> MetricDict:
            res = {}
            for m in self.metrics_selection:
                res = {**res, **m.metrics_step(self, preds, targets, **kwargs)}
            return res

        def metrics_epoch(
            self, preds: Tensor, targets: Tensor, prefix: str = "", *args, **kwargs
        ) -> MetricDict:
            res = {}
            for m in self.metrics_selection:
                res = {
                    **res,
                    **m.metrics_epoch(self, preds, targets, prefix, *args, **kwargs),
                }
            return res

    return MetricSelectorMixin


class MeanAveragePrecisionMetric(MetricMixin):
    """Mean Average Precision (mAP) metric"""

    def validate_attributes(self):
        for attribute in ["hparams.loss", "classes"]:
            attrgetter(attribute)(self)

    def _compute_mean_average_precision(self, preds, targets):
        try:
            ap = average_precision(
                preds,
                targets,
                num_classes=targets.shape[-1],
            )
        except RuntimeError as e:  # pragma: no cover
            logger.error("Unable to compute Average Precision: ", e)
            return torch.tensor(float("nan"))

        if isinstance(getattr(self, "ignore_classes", None), list):
            ap = [t for i, t in enumerate(ap) if i not in self.ignore_classes]

        if isinstance(ap, list):
            ap = torch.tensor([t for t in ap if not t.isnan()])

        return ap.mean()

    @classmethod
    def _metrics(cls):
        return {"mAP": OptimisationDirection.MAX}

    def metrics_step(
        self, preds: Tensor, targets: Tensor, *args, **kwargs
    ) -> MetricDict:
        return {"mAP": self._compute_mean_average_precision(preds, targets)}

    def metrics_epoch(
        self, preds: Tensor, targets: Tensor, *args, **kwargs
    ) -> MetricDict:
        return {"mAP": self._compute_mean_average_precision(preds, targets)}


def TopKAccuracyMetric(*Ks) -> MetricMixin:
    if not Ks:
        Ks = [1, 3, 5]

    for k in Ks:
        assert type(k) == int and k > 0

    class TopKAccuracyMetricClass(MetricMixin):
        """Top K accuracy metrics: top1acc, top3acc, top5acc"""

        @classmethod
        def _metrics(cls):
            return {f"top{k}acc": OptimisationDirection.MAX for k in Ks}

        def metrics_step(self, preds: Tensor, targets: Tensor, **kwargs) -> MetricDict:
            ks = [k for k in Ks]
            accs = [torch.tensor(-1.0) for _ in ks]
            try:
                accs = topk_accuracies(preds, targets, ks)
            except RuntimeError:  # pragma: no cover
                logger.error("Unable to compute top-k accuracy.")
            return {f"top{k}acc": accs[i] for i, k in enumerate(ks)}

    return TopKAccuracyMetricClass


class FlopsMetric(MetricMixin):
    """Computes Floating Point Operations (FLOPs) for the model and adds it as metric"""

    @classmethod
    def _metrics(cls):
        return {"flops": OptimisationDirection.MIN}

    def on_init_end(self, *args, **kwargs):
        assert isinstance(self, torch.nn.Module)
        self.flops = flops(self)  # type: ignore

    def metrics_step(self, preds: Tensor, targets: Tensor, **kwargs) -> MetricDict:
        return {"flops": torch.tensor(self.flops)}


class FlopsWeightedAccuracyMetric(FlopsMetric):
    """Computes acc * (flops / target_gflops) ** (-0.07)"""

    @classmethod
    def _metrics(cls):
        return {
            **{"flops_weighted_acc": OptimisationDirection.MAX},
            **FlopsMetric._metrics(),
        }

    def validate_attributes(self):
        for hparam in FlopsWeightedAccuracyMetric.configs().names:
            attrgetter(f"hparams.{hparam}")(self)

    @staticmethod
    def configs() -> Configs:
        c = Configs()
        c.add(
            name="target_gflops",
            type=float,
            default=2.0,
            strategy="constant",
            description="Target (Giga) Floating Point Operations per Second.",
        )
        return c

    def metrics_step(self, preds: Tensor, targets: Tensor, **kwargs) -> MetricDict:
        acc = topk_accuracies(preds, targets, ks=[1])[0]
        return {
            **FlopsMetric.metrics_step(self, preds, targets, **kwargs),
            "flops_weighted_acc": acc
            * (self.flops * 1e-9 / self.hparams.target_gflops) ** (-0.07),
        }


def topks_correct(preds: Tensor, labels: Tensor, ks: List[int]) -> List[Tensor]:
    """
    Given the predictions, labels, and a list of top-k values, compute the
    number of correct predictions for each top-k value.

    Args:
        preds (array): array of predictions. Dimension is batchsize
            N x ClassNum.
        labels (array): array of labels. Dimension is batchsize N.
        ks (list): list of top-k values. For example, ks = [1, 5] correspods
            to top-1 and top-5.

    Returns:
        topks_correct (list): list of numbers, where the `i`-th entry
            corresponds to the number of top-`ks[i]` correct predictions.
    """
    assert preds.size(0) == labels.size(
        0
    ), "Batch dim of predictions and labels must match"

    max_k = int(preds.shape[-1])

    # Find the top max_k predictions for each sample
    _, top_max_k_inds = torch.topk(preds, max_k, dim=1, largest=True, sorted=True)
    # (batch_size, max_k) -> (max_k, batch_size).
    top_max_k_inds = top_max_k_inds.t()
    # (batch_size, ) -> (max_k, batch_size).
    rep_max_k_labels = labels.view(1, -1).expand_as(top_max_k_inds)
    # (i, j) = 1 if top i-th prediction for the j-th sample is correct.
    top_max_k_correct = top_max_k_inds.eq(rep_max_k_labels)
    # Compute the number of topk correct predictions for each k.
    topks_correct: List[Tensor] = [
        top_max_k_correct[: min(k, max_k), :].reshape(-1).float().sum() for k in ks
    ]
    return topks_correct


def topk_errors(preds: Tensor, labels: Tensor, ks: List[int]):
    """
    Computes the top-k error for each k.
    Args:
        preds (array): array of predictions. Dimension is N.
        labels (array): array of labels. Dimension is N.
        ks (list): list of ks to calculate the top accuracies.
    """
    num_topks_correct = topks_correct(preds, labels, ks)
    return [(1.0 - x / preds.size(0)) for x in num_topks_correct]


def topk_accuracies(preds: Tensor, labels: Tensor, ks: List[int]):
    """
    Computes the top-k accuracy for each k.
    Args:
        preds (array): array of predictions. Dimension is N.
        labels (array): array of labels. Dimension is N.
        ks (list): list of ks to calculate the top accuracies.
    """
    num_topks_correct = topks_correct(preds, labels, ks)
    return [(x / preds.size(0)) for x in num_topks_correct]


def flops(model: torch.nn.Module):
    """Compute the Floating Point Operations per Second for the model"""
    return get_model_complexity_info(
        model,
        model.input_shape,
        as_strings=False,
        print_per_layer_stat=True,
        verbose=True,
    )[0]


def params_count(model: torch.nn.Module):
    """
    Compute the number of parameters.
    Args:
        model (model): model to count the number of parameters.
    """
    return np.sum([p.numel() for p in model.parameters()])


def make_confusion_matrix(
    preds: Tensor,
    targets: Tensor,
    classes: List[str],
) -> Figure:
    sns.set_theme()
    fig = plt.figure()
    z = (
        confusion_matrix(
            preds.argmax(1), targets, num_classes=len(classes), normalize="true"
        )
        .cpu()
        .numpy()
    )
    ax = sns.heatmap(
        z, annot=len(classes) <= 20, fmt=".2f", vmin=0, vmax=1, cmap="Blues"
    )
    for x, y in zip(ax.get_xticklabels(), ax.get_yticklabels()):
        x.set_text(f"{classes[int(x._text)]} ({x._text})")
        y.set_text(f"{classes[int(y._text)]} ({y._text})")
    ax.set_yticklabels(ax.get_yticklabels(), rotation=0)
    ax.set_xticklabels(ax.get_xticklabels(), rotation=90)
    plt.ylabel("True")
    plt.xlabel("Predicted")
    plt.tight_layout()
    return fig



import contextlib
from abc import abstractmethod
from time import time
from typing import Any, Dict, Tuple, overload

import numpy as np
import torch
from tqdm import tqdm

from ride.utils.gpus import parse_gpus, parse_num_gpus
from ride.utils.logging import getLogger
from ride.utils.utils import name, some

logger = getLogger(__name__)


@overload
def profile(model: torch.nn.Module, detailed=False) -> float:
    ...  # pragma: no cover


@overload
def profile(
    model: torch.nn.Module, detailed=True
) -> Tuple[float, torch.autograd.profiler.EventList,]:
    ...  # pragma: no cover


def profile(model: torch.nn.Module, detailed=True):
    for attr in [
        "device",
        "input_shape",
        "hparams.batch_size",
        "forward",
    ]:
        assert some(
            model, attr
        ), f"{name(model)} should define `{attr}` but none was found."

    prevously_training = model.training
    model.eval()

    # Move model to GPU if available
    prev_device = model.device
    gpus = parse_gpus(model.hparams.gpus) if hasattr(model.hparams, "gpus") else None
    new_device = f"cuda:{gpus[0]}" if gpus else "cpu"
    model.to(device=new_device)

    # Initialise data on CPU
    data = torch.randn(model.hparams.batch_size, *model.input_shape, device="cpu")
    model.warm_up(tuple(data.shape))

    try:
        with torch.no_grad(), torch.autograd.profiler.profile(
            use_cuda=(model.device.type != "cpu"), profile_memory=True
        ) if detailed else contextlib.suppress() as prof:
            # Transfer data from CPU to GPU, compute on GPU, and transfer back to CPU
            start = time()
            model(data.to(device=new_device)).to(device="cpu")
            stop = time()
    except Exception as e:
        raise ValueError(
            f"Caught exception while attempting to profile model. "
            "Did you ensure that you model has a correct `input_shape` attribute (omitting the batch_size dimension)? "
            f"Exception contents: {e}"
        )
    finally:
        model.to(device=prev_device)
        if prevously_training:
            model.train()

    timing = stop - start  # seconds

    if detailed:
        return timing, prof
    return timing


def compute_num_runs(total_wait_time, single_run_time):
    additional_runs = 1
    single_run_time = max(single_run_time, 1e-7)
    for option in [10000, 1000, 100, 10, 5, 3]:
        if (total_wait_time / single_run_time / option) > 1:
            additional_runs = option
            break
    return additional_runs


def profile_repeatedly(
    model: torch.nn.Module, max_wait_seconds=30, num_runs=None
) -> Tuple[Dict[str, str], torch.autograd.profiler.EventList]:
    for attr in ["hparams.batch_size"]:
        assert some(
            model, attr
        ), f"{name(model)} should define `{attr}` but none was found."

    # Make initial run to gauge its time. Also serves as a "dry-run"
    single_run_seconds, single_run_prof = profile(model, detailed=True)

    num_runs = (
        num_runs if num_runs else compute_num_runs(max_wait_seconds, single_run_seconds)
    )

    times_s = []
    for _ in tqdm(range(num_runs), desc="Profiling"):
        times_s.append(profile(model, detailed=False))

    batch_size = model.hparams.batch_size
    s_per_sample = np.array(times_s) / batch_size
    samples_per_s = 1 / s_per_sample
    µs_per_sample = s_per_sample * 1e6

    results_dict = {}
    results_dict[
        "time_per_sample"
    ] = f"{format_time(µs_per_sample.mean())} +/- {format_time(µs_per_sample.std())} [{format_time(µs_per_sample.min())}, {format_time(µs_per_sample.max())}]"
    results_dict[
        "samples_per_second"
    ] = f"{samples_per_s.mean():.3f} +/- {samples_per_s.std():.3f} [{samples_per_s.min():.3f}, {samples_per_s.max():.3f}]"
    results_dict["num_runs"] = num_runs
    results_dict["on_gpu"] = (
        parse_num_gpus(model.hparams.gpus) > 0
        if hasattr(model.hparams, "gpus")
        else False
    )
    results_dict["batch_size"] = batch_size

    return results_dict, single_run_prof


class ProfileableDataset:
    @abstractmethod
    def profile(self) -> Dict[str, Any]:
        ...  # pragma: no cover


def format_time(time_us):
    """Defines how to format time in FunctionEvent"""
    US_IN_SECOND = 1000.0 * 1000.0
    US_IN_MS = 1000.0
    if time_us >= US_IN_SECOND:
        return "{:.3f}s".format(time_us / US_IN_SECOND)
    if time_us >= US_IN_MS:
        return "{:.3f}ms".format(time_us / US_IN_MS)
    return "{:.3f}us".format(time_us)



import logging
import pickle
import shutil
from pathlib import Path
from typing import Tuple

import pytest
import torch

from ride.core import RideModule
from ride.finetune import Finetunable
from ride.optimizers import AdamWCyclicLrOptimizer
from ride.utils.checkpoints import get_latest_checkpoint
from ride.utils.utils import AttributeDict

# from ride.finetune import Finetunable
from .dummy_dataset import DummyRegressionDataLoader

from ride import Main, Configs  # noqa: F401  # isort:skip


class DummyModule(
    RideModule, Finetunable, DummyRegressionDataLoader, AdamWCyclicLrOptimizer
):
    def __init__(self, hparams):
        self.l1 = torch.nn.Linear(self.input_shape[0], self.hparams.hidden_dim)
        self.l2 = torch.nn.Linear(self.hparams.hidden_dim, self.output_shape)
        self.loss = torch.nn.functional.mse_loss

    def forward(self, x):
        x = x.view(x.size(0), -1)
        x = torch.relu(self.l1(x))
        x = torch.relu(self.l2(x))
        return x

    @staticmethod
    def configs():
        c = Configs()
        c.add(
            name="hidden_dim",
            type=int,
            default=128,
            strategy="choice",
            choices=[128, 256, 512, 1024],
            description="Number of hiden units.",
        )
        return c


def default_main_and_args() -> Tuple[Main, AttributeDict]:
    m = Main(DummyModule)
    parser = m.argparse(run=False)
    args, _ = parser.parse_known_args()
    args.max_epochs = 1
    args.gpus = 0
    args.checkpoint_callback = True
    args.optimization_metric = "loss"
    args.id = "automated_test"
    args.test_ensemble = 0
    args.checkpoint_every_n_steps = 0
    args.monitor_lr = 0
    args.auto_lr_find = 0
    args.auto_scale_batch_size = 0
    args.num_workers = 1
    args.batch_size = 4

    return m, args


@pytest.fixture()  # scope="module"
def main_and_args() -> Tuple[Main, AttributeDict]:
    return default_main_and_args()


class TestFinetuning:
    def test_finetune_from_weights(
        self, caplog, main_and_args: Tuple[Main, AttributeDict]
    ):
        """
        Test finetune_from_weights pickle and pyth suffixes
        """
        m, args = main_and_args
        args.train = True
        args.discriminative_lr_fraction = 0.5

        # Create a run to start with
        m.main(args)

        # Create checkpoints
        ckpt_cp = get_latest_checkpoint(m.log_dir)

        pth_cp = Path(m.log_dir) / "pretrained.pth"
        torch.save(m.runner.best_model.state_dict(), str(pth_cp))

        pth_cp_2 = Path(m.log_dir) / "pretrained_2.pth"
        torch.save(m.runner.best_model, str(pth_cp_2))

        pkl_cp = Path(m.log_dir) / "pretrained.pkl"
        with open(pkl_cp, "wb") as f:
            pickle.dump(m.runner.best_model.state_dict(), f)

        # Finetune from checkpoint
        for cp in [ckpt_cp, pth_cp, pth_cp_2, pkl_cp]:
            # for cp in [pth_cp]:
            m, args = default_main_and_args()  # Need to make new main
            args.finetune_from_weights = str(cp)
            args.test = True

            caplog.clear()
            with caplog.at_level(logging.INFO):
                m.main(args)

            assert len(caplog.messages) > 0
            assert any(
                [
                    f"Loading model weights from {str(cp)}" in msg
                    for msg in caplog.messages
                ]
            )

        # Clean up
        ckpt_cp.unlink()
        pth_cp.unlink()
        pkl_cp.unlink()

    def test_complex_finetuning(
        self, caplog, main_and_args: Tuple[Main, AttributeDict]
    ):
        """
        Test complex fine-tuning setup, including
        - finetune_from_weights
        - discriminative_lr_fraction
        - gradual_unfreeze
        """
        m, args = main_and_args
        args.train = True
        args.discriminative_lr_fraction = 0.5

        # Create a run to start with
        m.main(args)

        latest_checkpoint = get_latest_checkpoint(m.log_dir)
        assert latest_checkpoint.exists()
        checkpoint = Path("test_complex_finetuning.ckpt")
        shutil.copy(latest_checkpoint, checkpoint)

        # Finetune from checkpoint
        # Two epochs, both unfrozen
        m, args = default_main_and_args()  # Need to make new main
        args.finetune_from_weights = str(checkpoint)
        args.unfreeze_layers_initial = 1
        args.unfreeze_epoch_step = 1
        args.unfreeze_from_epoch = 0
        args.train = True
        args.max_epochs = 3

        caplog.clear()
        with caplog.at_level(logging.INFO):
            m.main(args)

        assert any(["Epoch 0, unfreezing 1 layer(s)" in msg for msg in caplog.messages])
        assert any(["Epoch 1, unfreezing 2 layer(s)" in msg for msg in caplog.messages])
        assert m.runner.trainer.model.l1.weight.requires_grad is True
        assert m.runner.trainer.model.l1.bias.requires_grad is True
        assert m.runner.trainer.model.l2.weight.requires_grad is True
        assert m.runner.trainer.model.l2.bias.requires_grad is True

        checkpoint.unlink()



import logging
import os
import sys
from pathlib import Path
from typing import Tuple

import pytest
import torch

from ride.core import RideModule
from ride.finetune import Finetunable
from ride.optimizers import AdamWOneCycleOptimizer
from ride.utils.checkpoints import get_latest_checkpoint
from ride.utils.io import dump_json, dump_yaml, is_nonempty_file
from ride.utils.utils import AttributeDict, attributedict, temporary_parameter

# from ride.finetune import Finetunable
from .dummy_dataset import DummyRegressionDataLoader

from ride import Main, Configs  # noqa: F401  # isort:skip


class DummyModule(
    RideModule, Finetunable, DummyRegressionDataLoader, AdamWOneCycleOptimizer
):
    def __init__(self, hparams):
        self.l1 = torch.nn.Linear(self.input_shape[0], self.hparams.hidden_dim)
        self.l2 = torch.nn.Linear(self.hparams.hidden_dim, self.output_shape)
        self.loss = torch.nn.functional.mse_loss

    def forward(self, x):
        x = x.view(x.size(0), -1)
        x = torch.relu(self.l1(x))
        x = torch.relu(self.l2(x))
        return x

    @staticmethod
    def configs():
        c = Configs()
        c.add(
            name="hidden_dim",
            type=int,
            default=128,
            strategy="choice",
            choices=[128, 256, 512, 1024],
            description="Number of hiden units.",
        )
        return c


def default_main_and_args() -> Tuple[Main, AttributeDict]:
    m = Main(DummyModule)
    parser = m.argparse(run=False)
    args, _ = parser.parse_known_args()
    args.max_epochs = 1
    args.gpus = 0
    args.checkpoint_callback = True
    args.optimization_metric = "loss"
    args.id = "automated_test"
    args.test_ensemble = 0
    args.checkpoint_every_n_steps = 0
    args.monitor_lr = 0
    args.auto_lr_find = 0
    args.auto_scale_batch_size = 0
    args.num_workers = 1
    args.batch_size = 4

    return m, args


@pytest.fixture()  # scope="module"
def main_and_args() -> Tuple[Main, AttributeDict]:
    return default_main_and_args()


# For unknown reasons, `test_train_val_test_combination` fails when run after the other tests
# in the suite. When running it first, everything works as expected...
@pytest.mark.order(1)
class TestMain:
    def test_help(self, capsys, caplog):
        """Test that command help works"""

        # Nothing is logged
        caplog.clear()
        with caplog.at_level(logging.WARNING):
            Main(DummyModule).argparse([])

        assert caplog.messages == [] or (
            len(caplog.messages) == 1 and "Missing log" in caplog.messages[0]
        )
        captured = capsys.readouterr()
        assert captured.out == ""
        assert captured.err == ""

        # Help is printed
        caplog.clear()
        with pytest.raises(SystemExit), caplog.at_level(logging.WARNING):
            Main(DummyModule).argparse(["--help"])

        # Help message was neither logged nor in stderr
        assert len(caplog.messages) == 0
        captured = capsys.readouterr()
        assert captured.err == ""

        # Help is in stdout
        help_msg = captured.out
        assert len(help_msg) > 0

        # Flow args
        for msg in [
            "--hparamsearch",
            "--train",
            "--test",
            "--profile_model",
        ]:
            assert msg in help_msg

        # General args
        for msg in ["--id", "--logging_backend", "--optimization_metric"]:
            assert msg in help_msg

        # Pytorch Lightning args
        for msg in [
            "--logger",
            "--gpus",
            "--accumulate_grad_batches",
            "--max_epochs",
            "--limit_train_batches",
            "--precision",
            "--resume_from_checkpoint",
            "--benchmark",
            "--auto_lr_find",
            "--auto_scale_batch_size",
        ]:
            assert msg in help_msg

        # Module args
        for msg in ["--loss", "--learning_rate", "--batch_size", "--hidden_dim"]:
            assert msg in help_msg

    def test_default_id(self):
        """Test that a default id is given"""
        m = Main(DummyModule)
        parser = m.argparse(run=False)
        args, _ = parser.parse_known_args()
        assert args.id == "unnamed"

    def test_main_arg_types(self, main_and_args: Tuple[Main, AttributeDict]):
        """Test if passing both dict and AttributeDict work for main"""
        m, args = main_and_args
        m.main(args)  # Namespace
        m.main(attributedict(args))  # AttributeDict
        m.main(vars(args))  # Dict

    def test_train_val_test_combination(
        self, caplog, main_and_args: Tuple[Main, AttributeDict]
    ):
        """Test that training, validation, and test works"""
        m, args = main_and_args
        args.train = True
        args.validate = True
        args.test = True
        args.checkpoint_every_n_steps = 10

        # Trainer args
        args.limit_val_batches = 2
        args.limit_test_batches = 2
        args.max_epochs = 2

        with caplog.at_level(logging.INFO):
            m.main(args)

        messages = "\n".join(caplog.messages)
        assert "Running training" in messages
        assert "val/loss" in messages
        assert "Epoch" in messages
        assert "Running evaluation on validation set" in messages
        assert "val/epoch" in messages
        assert "Running evaluation on test set" in messages
        assert "test/loss" in messages

        save_lines = [m for m in caplog.messages if "Saving" in m]

        # Check that result files exist
        hparams_path = Path(save_lines[0].split(" ")[2])
        assert is_nonempty_file(hparams_path)

        val_result_path = Path(save_lines[-2].split(" ")[2])
        assert is_nonempty_file(val_result_path)

        test_result_path = Path(save_lines[-1].split(" ")[2])
        assert is_nonempty_file(test_result_path)

        # Check that trainer args were passed
        assert m.runner.trainer.limit_val_batches == 2
        assert m.runner.trainer.limit_test_batches == 2

    def test_test_ensemble(self, caplog, main_and_args: Tuple[Main, AttributeDict]):
        """Test ensemble works"""
        m, args = main_and_args
        args.test = True
        args.test_ensemble = True

        with caplog.at_level(logging.INFO):
            m.main(args)

        for check in [
            "Running evaluation on test set using ensemble testing",
            "test/epoch",
        ]:
            assert any([check in msg for msg in caplog.messages])

    def test_from_hparams_file(self, caplog, main_and_args: Tuple[Main, AttributeDict]):
        """Test loading of hparams"""
        m, args = main_and_args
        args.validate = True

        for suffix, dump, hidden_dim in [
            ("yaml", dump_yaml, 256),
            ("json", dump_json, 512),
        ]:
            # Prep file
            hparams_path = Path(os.getcwd()) / f"test_dummy_module_hparams.{suffix}"
            dump(
                hparams_path,
                {"hidden_dim": hidden_dim, "batch_size": 4, "learning_rate": 0.01},
            )

            # Test
            caplog.clear()
            args.from_hparams_file = str(hparams_path)
            # args.train = True  # Triggers auto_scale_lr
            args.batch_size = 6
            with temporary_parameter(
                sys, "argv", [*sys.argv, "--batch_size", "6"]
            ), caplog.at_level(logging.INFO):
                m.main(args)

            assert len(caplog.messages) > 0
            for check in ["Scaling learning_rate from 0.01 to 0.015"]:
                any([check in msg for msg in caplog.messages])

            assert m.runner.trainer.model.hparams.hidden_dim == hidden_dim
            assert m.runner.trainer.model.hparams.batch_size == 6  # from command line
            assert m.runner.trainer.model.hparams.learning_rate == 0.015  # auto scaled

            # Cleanup
            hparams_path.unlink()

    def test_resume_from_checkpoint(self, main_and_args: Tuple[Main, AttributeDict]):
        """Test that resuming from checkpoint works with inferred checkpoint path works"""
        m, args = main_and_args
        args.train = True

        m.main(args)

        # From specific file (as usual in PyTorch Lightning)
        args.resume_from_checkpoint = str(get_latest_checkpoint(m.log_dir))
        assert ".ckpt" in args.resume_from_checkpoint
        m.main(args)  # Doesn't run because max_epochs was reached in checkpoint

        # Automatically pick lastest checkpoint in directory
        args.resume_from_checkpoint = m.log_dir
        assert ".ckpt" not in args.resume_from_checkpoint
        m.main(args)

    def test_profile_model(self, caplog, main_and_args: Tuple[Main, AttributeDict]):
        """Test that profiling model works"""
        m, args = main_and_args
        args.profile_model = True
        args.profile_model_num_runs = 10

        caplog.clear()

        # Nothing is logged
        with caplog.at_level(logging.INFO):
            m.main(args)

        assert len(caplog.messages) > 0

        for check in [
            "Results",
            "flops",
            "machine",
            "samples_per_second",
            "num_runs",
        ]:
            assert any([check in msg for msg in caplog.messages])

        # Check that results are saved and path is printed
        model_profile_path = Path(caplog.messages[-1].split(" ")[2])
        assert is_nonempty_file(model_profile_path)

    def test_hparamsearch(self, main_and_args: Tuple[Main, AttributeDict]):
        """Test that execution of hparamsearch and retreival of best hparams works"""

        # Prepare hparamspace file
        hparams_space_path = Path(os.getcwd()) / "test_dummy_module_hparamspace.yaml"
        hparams_space = {
            "hidden_dim": {
                "type": "int",
                "strategy": "choice",
                "choices": [128, 256, 512, 1024],
            }
        }
        dump_yaml(hparams_space_path, hparams_space)

        # Prep args
        m, args = main_and_args
        args.hparamsearch = True
        args.max_epochs = 1
        args.trials = 2
        args.from_hparam_space_file = str(hparams_space_path)

        # Test
        m.main(args)

        # Clean up
        hparams_space_path.unlink()



import logging
from argparse import ArgumentParser

import pytest
import pytorch_lightning as pl
import torch

from ride.core import AttributeDict, RideMixin, RideModule
from ride.optimizers import SgdOptimizer
from ride.utils.utils import DictLike

from .dummy_dataset import DummyRegressionDataLoader

pl.seed_everything(42)


class Mixin1(RideMixin):
    def __init__(self, hparams: AttributeDict):
        self.hparams.msg.append("Mixin1.__init__")

    def on_init_end(self, hparams, *args, **kwargs):
        self.hparams.msg.append("Mixin1.on_init_end")


class Mixin2(RideMixin):
    def __init__(self, hparams: AttributeDict):
        self.hparams.msg.append("Mixin2.__init__")

    def on_init_end(self, hparams, *args, **kwargs):
        self.hparams.msg.append("Mixin2.on_init_end")


class InitOrderModule(
    RideModule,
    Mixin1,
    Mixin2,
):
    def __init__(self, hparams: DictLike):
        self.hparams.msg.append("InitOrderModule.__init__")
        self.input_shape = (1,)
        self.output_shape = (1,)


def test_init_order():

    module = InitOrderModule({"msg": []})

    assert module.hparams.msg == [
        "Mixin1.__init__",
        "Mixin2.__init__",
        "InitOrderModule.__init__",
        "Mixin1.on_init_end",
        "Mixin2.on_init_end",
    ]


class DummyModule(
    RideModule,
    SgdOptimizer,
    DummyRegressionDataLoader,
):
    # Not needed:
    # @staticmethod
    # def configs() -> Configs:
    #     return Configs.collect(DummyModule)

    def __init__(self):
        self.lin = torch.nn.Linear(
            self.input_shape[0],  # from DummyRegressionDataLoader
            self.output_shape,  # from DummyRegressionDataLoader
        )
        # Alternative way of specifying loss:
        self.loss = torch.nn.functional.mse_loss

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.lin(x)


def test_init_only_self():
    """Test if modules initialises if"""
    DummyModule()


# From pt v1.4 training_step looks for module.trainer, which this test doesn't take into account
def xtest_training_step_added_automatically():
    """Test if modules adds lifecycle automatically"""

    parser = DummyModule.configs().add_argparse_args(ArgumentParser())
    args, _ = parser.parse_known_args()
    args = apply_standard_args(args)
    module = DummyModule(args)
    assert module.loss.__name__ == "mse_loss"
    batch_size = 2
    x = torch.ones((batch_size, *module.input_shape), dtype=torch.float)
    y = torch.ones((batch_size, module.output_shape), dtype=torch.float)
    step_output = module.training_step((x, y), batch_idx=0)

    assert isinstance(step_output["loss"], torch.Tensor)
    assert step_output["pred"].shape == y.shape
    assert torch.equal(step_output["target"], y)


def apply_standard_args(args):
    args.max_epochs = 3
    args.gpus = 0
    args.checkpoint_callback = True
    args.performance_metric = "loss"
    args.id = "automated_test"
    args.test_ensemble = 0
    args.loss = "mse_loss"
    return args


def test_module_with_no_forward_warns(caplog):
    """Test if modules initialises when hparams is also given"""
    caplog.clear()
    with caplog.at_level(logging.WARNING):

        class DummyModuleNoForward(
            RideModule,
            SgdOptimizer,
            DummyRegressionDataLoader,
        ):
            def __init__(self):
                self.lin = torch.nn.Linear(
                    self.input_shape[0],  # from DummyRegressionDataLoader
                    self.output_shape,  # from DummyRegressionDataLoader
                )

            # Missing on purpose:
            # def forward(self, x: torch.Tensor) -> torch.Tensor:
            #     return self.lin(x)

    assert len(caplog.messages) == 1
    assert "forward" in caplog.text


@pytest.mark.skip(reason="Nested inheritance is not supported")
def test_child_model():
    """Test if inheritance from RideModule works"""

    class DummyChild(DummyModule):
        def __init__(self, hparams):
            DummyModule.__init__(self, hparams)
            self.some_variable = 42

    parser = DummyChild.configs().add_argparse_args(ArgumentParser())
    args, _ = parser.parse_known_args()
    args = apply_standard_args(args)
    module = DummyChild(args)
    assert module.some_variable == 42
    assert module.loss.__name__ == "mse_loss"
    batch_size = 2
    x = torch.ones((batch_size, *module.input_shape), dtype=torch.float)
    y = torch.ones((batch_size, module.output_shape), dtype=torch.float)
    step_output = module.training_step((x, y), batch_idx=0)

    assert isinstance(step_output["loss"], torch.Tensor)
    assert step_output["pred"].shape == y.shape
    assert torch.equal(step_output["target"], y)


def test_default_methods_overload():
    # RideMixins can overload default_methods, e.g. `warm_up`

    msg = None

    class MyWarmup(RideMixin):
        def warm_up(self, input_shape, *args, **kwargs):
            nonlocal msg
            msg = input_shape

    class DummyModuleWithWarmup(RideModule, MyWarmup, DummyRegressionDataLoader):
        def __init__(self):
            self.lin = torch.nn.Linear(
                self.input_shape[0],  # from DummyRegressionDataLoader
                self.output_shape,  # from DummyRegressionDataLoader
            )

        def forward(self, x):
            return self.lin(x)

    parser = DummyModuleWithWarmup.configs().add_argparse_args(ArgumentParser())
    args, _ = parser.parse_known_args()
    args = apply_standard_args(args)
    module = DummyModuleWithWarmup(args)

    assert msg is None

    module.warm_up((1, 2, 3))

    assert msg == (1, 2, 3)



import os
from pathlib import Path

from dotenv import find_dotenv, load_dotenv

load_dotenv(find_dotenv(usecwd=True))

__all__ = [
    "DATASETS_PATH",
    "LOGS_PATH",
    "RUN_LOGS_PATH",
    "TUNE_LOGS_PATH",
    "CACHE_PATH",
    "LOG_LEVEL",
    "NUM_CPU",
]

DATASETS_PATH = Path(os.getenv("DATASETS_PATH", default="datasets"))
LOGS_PATH = Path(os.getenv("LOGS_PATH", default="logs"))
RUN_LOGS_PATH = LOGS_PATH / "run_logs"
TUNE_LOGS_PATH = LOGS_PATH / "tune_logs"
CACHE_PATH = Path(os.getenv("CACHE_PATH", default=".cache"))
LOG_LEVEL = os.getenv("LOG_LEVEL", default="INFO")
NUM_CPU = os.cpu_count() or 1

DATASETS_PATH.mkdir(exist_ok=True)
LOGS_PATH.mkdir(exist_ok=True)
RUN_LOGS_PATH.mkdir(exist_ok=True)
TUNE_LOGS_PATH.mkdir(exist_ok=True)
CACHE_PATH.mkdir(exist_ok=True)



import os
from argparse import Namespace
from pathlib import Path
from typing import Any, Callable, Dict, List, Type

from ptflops import get_model_complexity_info
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import LearningRateMonitor, ModelCheckpoint
from pytorch_lightning.core.lightning import LightningModule
from pytorch_lightning.utilities.cloud_io import load as pl_load
from pytorch_lightning.utilities.parsing import AttributeDict

from ride.core import RideDataset, RideModule
from ride.logging import (
    ExperimentLoggerCreator,
    ResultsLogger,
    add_experiment_logger,
    experiment_logger,
)
from ride.profile import profile_repeatedly
from ride.utils.logging import getLogger, process_rank
from ride.utils.machine_info import get_machine_info
from ride.utils.utils import attributedict

EvalutationResults = Dict[str, float]

logger = getLogger(__name__)


def is_runnable(cls):
    assert issubclass(cls, RideModule)
    assert issubclass(cls, RideDataset)
    assert issubclass(cls, LightningModule)
    return True


class Runner:
    trained_model: RideModule

    def __init__(
        self,
        Module: Type[RideModule],
    ):
        assert is_runnable(Module)
        self.Module = Module

    def train(
        self,
        args: AttributeDict,
        trainer_callbacks: List[Callable] = [],
        tune_checkpoint_dir: str = None,
        experiment_logger: ExperimentLoggerCreator = experiment_logger,
    ) -> RideModule:
        args = attributedict(args)

        # Support for checkpoint loading in Tune
        # c.f. https://docs.ray.io/en/master/tune/tutorials/tune-pytorch-lightning.html#adding-checkpoints-to-the-pytorch-lightning-module
        if tune_checkpoint_dir:
            tune_ckpt = pl_load(
                os.path.join(tune_checkpoint_dir, "checkpoint"),
                map_location=lambda storage, loc: storage,
            )
            model = self.Module._load_model_state(tune_ckpt, hparams=args)

        else:
            model = self.Module(hparams=args)

        # Handle distributed backend modes in Lightning
        if not getattr(args, "distributed_backend", None):
            if model.hparams.num_gpus > 1:
                args.distributed_backend = "ddp"
            else:
                args.distributed_backend = None

        # Ensure callbacks are not stacked if multiple calls are made
        trainer_callbacks = trainer_callbacks.copy()

        # Prepare logger and callbacks
        trainer_callbacks.append(
            ModelCheckpoint(
                save_top_k=1,
                verbose=True,
                monitor=f"val/{args.optimization_metric}",  # Comment out when using pl.EvalResult
                mode=self.Module.metrics()[args.optimization_metric].value,
                save_last=True,
                every_n_train_steps=args.checkpoint_every_n_steps or None,
            )
        )
        logger.info(
            f"✅ Checkpointing on val/{args.optimization_metric} with optimisation direction {self.Module.metrics()[args.optimization_metric].value}"
        )
        if args.monitor_lr:
            trainer_callbacks.append(LearningRateMonitor(logging_interval="step"))

        _experiment_logger = experiment_logger(args.id, args.logging_backend)

        self.trainer = Trainer.from_argparse_args(
            Namespace(**args),
            logger=_experiment_logger,
            callbacks=trainer_callbacks,
        )
        _experiment_logger.log_hyperparams(dict(**model.hparams))

        # Load epoch state for Tune checkpoint
        if tune_checkpoint_dir:
            self.trainer.current_epoch = tune_ckpt["epoch"]  # type:ignore

        # Run hparam routines in Lightning
        if args.auto_scale_batch_size:
            # with temporary_parameter(self.trainer, "auto_lr_find", 0):
            self.trainer.tune(model)

        if args.auto_lr_find:
            lr_finder = self.trainer.tuner.lr_find(model, min_lr=1e-8, max_lr=1e-1)
            lr_suggestion = lr_finder.suggestion()
            logger.info(f"Suggested learning rate is {lr_suggestion}")
            model.hparams.learning_rate = lr_suggestion

            # Plot suggestion
            if process_rank == 0:
                fig = lr_finder.plot(suggest=True)
                lr_fig_path = Path(_experiment_logger.log_dir) / "lr_find.png"
                logger.info(f"Saving plot of learning rate sweep to {lr_fig_path}")
                lr_fig_path.parent.mkdir(parents=True, exist_ok=True)
                fig.savefig(lr_fig_path)

        self.trainer.fit(model)

        if Path(self.trainer.checkpoint_callback.best_model_path).is_file():
            self.best_model = self.Module.load_from_checkpoint(
                checkpoint_path=self.trainer.checkpoint_callback.best_model_path,
                hparams=model.hparams,
            )

        return model

    def evaluate(self, args: AttributeDict, mode="val") -> EvalutationResults:
        assert mode in {"val", "test"}
        args = attributedict(args)

        # Init model
        if hasattr(self, "best_model"):
            model = self.best_model
        else:
            model = self.Module(hparams=args)

        dataloaders = {"val": model.val_dataloader, "test": model.test_dataloader}

        # Init trainer
        base_logger = experiment_logger(args.id, args.logging_backend)
        results_logger = ResultsLogger(prefix=mode, save_to=base_logger.log_dir)
        logger = add_experiment_logger(base_logger, results_logger)
        if hasattr(self, "trainer"):
            self.trainer.logger = logger
        else:
            self.trainer = Trainer.from_argparse_args(Namespace(**args), logger=logger)

        base_logger.log_hyperparams(dict(**model.hparams))
        # Run eval
        self.trainer.test(model, dataloaders[mode]())

        results = results_logger.results

        # It seems that `trainer.test` actively sets the model dataloader; revert it
        model.val_dataloader = dataloaders["val"]
        model.test_dataloader = dataloaders["test"]

        return results

    def validate(self, args: AttributeDict) -> EvalutationResults:
        return self.evaluate(args, "val")

    def test(self, args: AttributeDict) -> EvalutationResults:
        return self.evaluate(args, "test")

    def train_and_val(
        self,
        args: AttributeDict,
        trainer_callbacks: List[Callable] = [],
        tune_checkpoint_dir: str = None,
        experiment_logger: ExperimentLoggerCreator = experiment_logger,
    ) -> EvalutationResults:
        self.train(args, trainer_callbacks, tune_checkpoint_dir, experiment_logger)
        return self.evaluate(args, mode="val")

    @staticmethod
    def static_train_and_val(
        Module: Type[RideModule],
        args: AttributeDict,
        trainer_callbacks: List[Callable] = [],
        tune_checkpoint_dir: str = None,
        experiment_logger: ExperimentLoggerCreator = experiment_logger,
    ) -> EvalutationResults:
        return Runner(Module).train_and_val(
            args, trainer_callbacks, tune_checkpoint_dir, experiment_logger
        )

    def profile_model(
        self, args: AttributeDict, max_wait_seconds: float = 10, num_runs: int = None
    ) -> Dict[str, Any]:
        if hasattr(self, "trained_model"):
            model = self.trained_model
        else:
            model = self.Module(hparams=args)

        timing_results_dict, single_profile = profile_repeatedly(
            model, max_wait_seconds, num_runs
        )

        single_run_detailed_timing = single_profile.key_averages().table(
            sort_by="self_cpu_time_total"
        )
        logger.info(single_run_detailed_timing)

        model.warm_up((1, *model.input_shape))

        flops, params = get_model_complexity_info(
            model,
            model.input_shape,
            as_strings=False,
            print_per_layer_stat=True,
            verbose=True,
        )

        elogger = experiment_logger(args.id, args.logging_backend)
        elogger.log_hyperparams(dict(**model.hparams))
        elogger.log_metrics(
            {
                "flops": int(flops),
                "params": int(params),
            }
        )

        return {
            "timing": timing_results_dict,
            "flops": int(flops),
            "params": int(params),
            "machine": get_machine_info(),
        }

    def find_learning_rate(self):
        raise NotImplementedError()

    def find_batch_size(self):
        raise NotImplementedError()



from ride import info


def test_info():
    assert type(info.__version__) == str
    assert type(info.__author__) == str
    assert type(info.__author_email__) == str
    assert type(info.__license__) == str
    assert type(info.__copyright__) == str
    assert type(info.__homepage__) == str
    assert type(info.__docs__) == str



from .main import Main  # noqa: F401, E402  # isort:skip
from .core import (  # noqa: F401, E402
    Configs,
    RideClassificationDataset,
    RideDataset,
    RideModule,
    getLogger,
)
from .finetune import Finetunable  # noqa: F401
from .hparamsearch import Hparamsearch  # noqa: F401, E402
from .lifecycle import Lifecycle  # noqa: F401, E402
from .metrics import (  # noqa: F401, E402
    FlopsMetric,
    FlopsWeightedAccuracyMetric,
    MeanAveragePrecisionMetric,
    MetricSelector,
    TopKAccuracyMetric,
)
from .optimizers import (  # noqa: F401, E402
    AdamWOneCycleOptimizer,
    AdamWOptimizer,
    SgdOneCycleOptimizer,
    SgdOptimizer,
)



from itertools import groupby
from operator import attrgetter
from typing import Callable, Dict

import torch
import torch.nn.functional as F
from pytorch_lightning import LightningModule
from supers import supers

from ride.core import Configs
from ride.logging import log_figures
from ride.metrics import (
    MetricDict,
    MetricMixin,
    OptimisationDirection,
    sort_out_figures,
)
from ride.utils.gpus import parse_num_gpus
from ride.utils.logging import getLogger
from ride.utils.utils import some_callable

loss_names = {
    "binary_cross_entropy",
    "binary_cross_entropy_with_logits",
    "poisson_nll_loss",
    # cosine_embedding_loss(input1, input2, target)
    "cross_entropy",
    # ctc_loss(log_probs, targets, input_lengths, target_lengths)
    "hinge_embedding_loss",
    "kl_div",
    "l1_loss",
    "mse_loss",
    # margin_ranking_loss(input1, input2, target)
    "multilabel_margin_loss",
    "multilabel_soft_margin_loss",
    "multi_margin_loss",
    "nll_loss",
    "smooth_l1_loss",
    "soft_margin_loss",
    # triplet_margin_loss(anchor, positive, negative)
    # triplet_margin_with_distance_loss(anchor, positive, negative)
}

logger = getLogger(__name__)


class Lifecycle(MetricMixin):
    """Adds train, val, and test lifecycle methods with cross_entropy loss

    During its `traning_epoch_end(epoch)` lifecycle method,
    it will call `on_traning_epoch_end` for all superclasses of
    its child class
    """

    hparams: ...
    forward: Callable[[torch.Tensor], torch.Tensor]
    _epoch: int

    @classmethod
    def _metrics(cls):
        return {"loss": OptimisationDirection.MIN}

    def __init__(self, hparams=None, *args, **kwargs):
        self._epoch = 0
        self.hparams.num_gpus = parse_num_gpus(hparams.gpus)
        self._sync_dist = self.hparams.num_gpus > 1
        if not some_callable(self, "loss", min_num_args=2):
            self.loss = attrgetter(hparams.loss)(F)

    def validate_attributes(self):
        for attribute in [
            "forward",
            "training_step",
            "validation_step",
            "test_step",
        ]:
            attrgetter(attribute)(self)

        for hparam in Lifecycle.configs().names:
            attrgetter(f"hparams.{hparam}")(self)

    @staticmethod
    def configs() -> Configs:
        c = Configs()
        c.add(
            name="optimization_metric",
            default="loss",
            type=str,
            choices=MetricMixin.metric_names(),
            description="Name of the performance metric that should be optimized",
        )
        c.add(
            name="test_ensemble",
            type=int,
            default=0,
            strategy="constant",
            description="Flag indicating whether the test dataset should yield a clip ensemble.",
        )
        c.add(
            name="gpus",
            type=str,
            default=None,
            strategy="constant",
            description="Which gpus should be used. Can be either the number of gpus (e.g. '2') or a list of gpus (e.g. ('0,1').",
        )
        c.add(
            name="loss",
            type=str,
            default="cross_entropy",
            choices=loss_names,
            strategy="constant",
            description="Loss function used during optimisation.",
        )
        return c

    def metrics_step(
        self, preds: torch.Tensor, targets: torch.Tensor, **kwargs
    ) -> MetricDict:
        return {"loss": self.loss(preds, targets)}

    def common_step(self, pred, target, prefix="train/", log=False):
        opt_key = prefix + self.hparams.optimization_metric
        loss_key = prefix + "loss"
        metrics = prefix_keys(prefix, self.collect_metrics(pred, target))

        if log:
            LightningModule.log(
                self,
                name=f"step_{loss_key}",
                value=metrics[loss_key],
                prog_bar=True,
                logger=True,
                sync_dist=self._sync_dist,
            )
            if opt_key != loss_key and opt_key in metrics:
                LightningModule.log(
                    self,
                    name=f"step_{opt_key}",
                    value=metrics[opt_key],
                    prog_bar=True,
                    logger=True,
                    sync_dist=self._sync_dist,
                )
        return {
            "loss": metrics[loss_key],
            **{
                k: v.detach() if hasattr(v, "detach") else v for k, v in metrics.items()
            },
            "pred": pred.detach().cpu(),
            "target": target.detach().cpu(),
        }

    def common_epoch_end(
        self, step_outputs, prefix="train/", exclude_keys={"pred", "target"}
    ):
        keys = list(step_outputs[0].keys())
        mean_step_metrics = {
            k: torch.mean(torch.stack([x[k] for x in step_outputs]))
            for k in keys
            if k not in exclude_keys
        }
        preds, targets = zip(*[(s["pred"], s["target"]) for s in step_outputs])
        preds = torch.vstack([p.unsqueeze(-1) for p in preds]).squeeze(-1)
        targets = torch.vstack([t.unsqueeze(-1) for t in targets]).squeeze(-1)
        epoch_metrics = prefix_keys(
            prefix,
            self.collect_epoch_metrics(preds, targets, prefix.replace("/", "")),
        )
        epoch_metrics, epoch_figures = sort_out_figures(epoch_metrics)
        all_metrics = {**mean_step_metrics, **epoch_metrics}
        LightningModule.log_dict(self, all_metrics, sync_dist=self._sync_dist)
        log_figures(self, epoch_figures)

    def preprocess_batch(self, batch):
        return batch

    def training_step(self, batch, batch_idx=None):
        if batch_idx == 0:
            supers(self).on_traning_epoch_start(self._epoch)
        batch = self.preprocess_batch(batch)
        x, target = batch[0], batch[1]
        pred = self.forward(x)
        return self.common_step(pred, target, prefix="train/", log=True)

    def training_epoch_end(self, step_outputs):
        self._epoch += 1
        self.common_epoch_end(step_outputs, prefix="train/")

    def validation_step(self, batch, batch_idx=None):
        batch = self.preprocess_batch(batch)
        x, target = batch[0], batch[1]
        pred = self.forward(x)
        return self.common_step(pred, target, prefix="val/")

    def validation_epoch_end(self, step_outputs):
        self.common_epoch_end(step_outputs, prefix="val/")

    def test_step(self, batch, batch_idx=None):
        if batch is None:
            return None

        batch = self.preprocess_batch(batch)
        x, target = batch[0], batch[1]
        pred = self.forward(x)

        if not self.hparams.test_ensemble:
            return {
                **self.common_step(pred, target, prefix="test/"),
                "pred": pred,
                "target": target,
            }

        identifier = batch[-1]
        # Delay computation of metrics to epoch end
        return {"pred": pred, "target": target, "identifier": identifier}

    def test_epoch_end(self, step_outputs):
        if self.hparams.test_ensemble:
            keys = set(step_outputs[0].keys())
            assert keys == {"pred", "target", "identifier"}

            # Unbatching step_outputs
            step_outputs = [
                {k: batch_step[k][i] for k in keys}
                for batch_step in step_outputs
                for i in range(len(batch_step["identifier"]))
            ]

            # Grouping samples by identifier
            steps_by_key = [
                (i, list(group))
                for i, group in groupby(
                    step_outputs, key=lambda x: int(x["identifier"])
                )
            ]

            # Averaging predictions
            preds_and_targets = [
                (
                    torch.mean(
                        torch.stack([s["pred"] for s in steps]), dim=0
                    ).unsqueeze(0),
                    steps[0]["target"].unsqueeze(0),  # unqueeze to add batch dim
                )
                for _, steps in steps_by_key
            ]

            # Computing steps
            step_outputs = [
                {
                    **self.common_step(pred, target, prefix="test/"),
                    "pred": pred,
                    "target": target,
                }
                for pred, target in preds_and_targets
            ]

        self.common_epoch_end(step_outputs, prefix="test/")


def prefix_keys(prefix: str, dictionary: Dict) -> Dict:
    return {f"{prefix}{k}": v for k, v in dictionary.items()}



import pickle
import re
from argparse import ArgumentError
from operator import attrgetter
from pathlib import Path

import pytorch_lightning as pl
import torch

from ride.core import Configs
from ride.unfreeze import Unfreezable
from ride.utils.logging import getLogger
from ride.utils.utils import attributedict, to_dict

logger = getLogger(__name__)


class Finetunable(Unfreezable):
    """Adds finetune capabilities to model

    This module is automatically added when RideModule is inherited
    """

    hparams: ...

    @staticmethod
    def configs() -> Configs:
        c: Configs = Unfreezable.configs()
        c.add(
            name="finetune_from_weights",
            default="",
            type=str,
            description=(
                "Path to weights to finetune from. "
                "Allowed extension include {'.ckpt', '.pyth', '.pth', '.pkl', '.pickle'}."
            ),
        )
        c.add(
            name="finetune_from_weights_key",
            default="",
            type=str,
            description="Key in weights-file, which should contains model state_dict in case of '.pyth' or '.pth' files.",
        )
        c.add(
            name="finetune_params_skip",
            default="",
            type=str,
            description=(
                "Regex for matching parameter names between fintune source and target model."
                "The parameter is not copied if `finetune_params_skip` is in paramer name."
            ),
        )
        return c

    def validate_attributes(self):
        for hparam in Finetunable.configs().names:
            attrgetter(f"hparams.{hparam}")(self)
        Unfreezable.validate_attributes(self)

    def map_loaded_weights(self, file, loaded_state_dict):
        return loaded_state_dict

    def on_init_end(self, hparams, *args, **kwargs):
        self.hparams.finetune_params_skip = (
            f".*({self.hparams.finetune_params_skip}).*"
            if self.hparams.finetune_params_skip
            else None
        )

        if not self.hparams.finetune_from_weights:
            Unfreezable.on_init_end(self, hparams, *args, **kwargs)
            return

        # Load model
        new_model_state = self.state_dict()  # type: ignore

        # Load hparams
        default_ft_hparams = to_dict(Finetunable.configs().default_values())
        hparams_passed = attributedict(
            {
                k: (default_ft_hparams[k] if k in default_ft_hparams else v)
                for k, v in self.hparams.items()
            }
        )

        state_dict = load_model_weights(
            self.hparams.finetune_from_weights,
            hparams_passed,
            self.hparams.finetune_from_weights_key,
        )
        state_dict = self.map_loaded_weights(
            self.hparams.finetune_from_weights, state_dict
        )

        def key_ok(k):
            return (
                k in new_model_state
                and k in state_dict
                and not (
                    self.hparams.finetune_params_skip
                    and re.match(self.hparams.finetune_params_skip, k)
                )
            )

        def size_ok(k):
            return new_model_state[k].size() == state_dict[k].size()

        to_load = {k: v for k, v in state_dict.items() if key_ok(k) and size_ok(k)}

        self.load_state_dict(to_load, strict=False)  # type: ignore

        names_missing = set(new_model_state.keys()) - set(to_load.keys())
        names_missing = {n for n in names_missing if "num_batches_tracked" not in n}

        if names_missing:
            logger.debug(f"missing keys: {sorted(names_missing)}")

        names_unexpected = set(state_dict.keys()) - set(new_model_state.keys())
        names_unexpected = {
            n for n in names_unexpected if "num_batches_tracked" not in n
        }

        if names_unexpected:
            logger.debug(f"unexpected keys: {sorted(names_unexpected)}")

        Unfreezable.on_init_end(
            self, hparams, names_to_unfreeze=names_missing, *args, **kwargs
        )


def load_model_weights(file: str, hparams_passed, model_state_key):
    path = Path(file)
    suffix = path.suffix
    assert (
        path.exists()
    ), f"Unable to load model weights from non-existing file ({file})"
    logger.info(f"Loading model weights from {path}")

    if suffix in {".ckpt"}:
        return pl.utilities.cloud_io.load(
            file, map_location=lambda storage, loc: storage
        )["state_dict"]
    elif suffix in {".pyth", ".pth", ".pt"}:
        return try_pyth_load(file, model_state_key)
    elif suffix in {".pkl", ".pickle"}:
        return try_pickle_load(file)
    else:
        raise ArgumentError(
            f"Unable to load model weights with suffix '{suffix}'. Suffix must be one of {'.ckpt', '.pyth', '.pth', '.pt', '.pkl', '.pickle'}"
        )


def try_pyth_load(file, model_state_key):
    loaded = torch.load(file, map_location="cpu")
    if issubclass(type(loaded), torch.nn.Module):
        return loaded.state_dict()

    assert issubclass(
        type(loaded), dict
    ), "pyth checkpoint should either be a model or a dict of weights"

    guesses = [
        model_state_key,
        "state_dict",
        "model_state",
    ]
    for g in guesses:
        if g in loaded.keys():
            state_dict = loaded[g]
            break

    # Check if we already have a state_dict
    if all([isinstance(v, torch.Tensor) for v in loaded.values()]):
        state_dict = loaded

    if not state_dict:  # pragma: no cover
        raise KeyError(
            f"None of the tried keys {guesses} fits loaded model state {loaded.keys()}. You can try another key using the `finetune_from_weights_key` hparam."
        )

    return state_dict


def try_pickle_load(file):
    with open(file, "r") as f:
        try:
            return pickle.load(file)
        except Exception:
            pass

        try:
            return pickle.load(f, encoding="latin1")
        except Exception:
            pass

    with open(file, "rb") as f:
        try:
            return pickle.load(file)
        except Exception:
            pass

        try:
            return pickle.load(f, encoding="latin1")
        except Exception:  # pragma: no cover
            pass

    raise ValueError(f"Unable to load file {file}")  # pragma: no cover



import logging

import pytest
import torch
from torchmetrics.functional.classification import average_precision

from ride import metrics
from ride.core import RideModule
from ride.utils.utils import attributedict

from .dummy_dataset import DummyClassificationDataLoader


def test_faulty_metric_warning(caplog):
    """Test that base class has sensible defaults"""

    caplog.clear()
    with caplog.at_level(logging.ERROR):

        class MyMetric(metrics.MetricMixin):
            ...

    assert caplog.messages == [
        "Subclasses of MetricMixin should define a `_metrics` classmethod, but none was found in MyMetric"
    ]


# For unknown reasons, this test fails on the Github Actions build system
@pytest.mark.skip_cicd
def test_MeanAveragePrecisionMetric():
    class DummyModule(
        RideModule, metrics.MeanAveragePrecisionMetric, DummyClassificationDataLoader
    ):
        def __init__(self, hparams):
            self.lin = torch.nn.Linear(self.input_shape[0], self.output_shape)
            self.loss = torch.nn.functional.mse_loss

        def forward(self, x):
            x = x.view(x.size(0), -1)
            x = torch.relu(self.lin(x))
            return x

    targets = torch.tensor(
        [[0, 1], [0, 1], [0, 1], [0, 1], [1, 0], [1, 0], [1, 0], [1, 0]]
    )
    preds = torch.tensor([[0.1, 0.9] for _ in range(8)]).clone()

    pl_map = average_precision(preds, targets, num_classes=2)

    net = DummyModule()
    assert DummyModule.metric_names() == ["loss", "mAP"]
    assert net.metrics_step(preds, targets)["mAP"] == pl_map
    assert net.metrics_epoch(preds, targets)["mAP"] == pl_map


def test_TopKAccuracyMetric():
    class DummyModule(
        RideModule, DummyClassificationDataLoader, metrics.TopKAccuracyMetric()
    ):
        def __init__(self, hparams):
            self.lin = torch.nn.Linear(self.input_shape[0], self.output_shape)
            self.loss = torch.nn.functional.mse_loss

        def forward(self, x):
            x = x.view(x.size(0), -1)
            x = torch.relu(self.lin(x))
            return x

    net = DummyModule()
    targets = torch.tensor([[0], [0], [0], [0], [1], [1], [1], [1]])
    preds = torch.tensor([[0.1, 0.9] for _ in range(8)])

    assert DummyModule.metric_names() == ["loss", "top1acc", "top3acc", "top5acc"]
    assert net.metrics_step(preds, targets)["top1acc"] == torch.tensor(0.5)
    assert net.metrics_step(preds, targets)["top3acc"] == torch.tensor(1.0)
    assert net.metrics_step(preds, targets)["top5acc"] == torch.tensor(1.0)


def test_FlopsWeightedAccuracyMetric():
    class DummyModule(
        RideModule, DummyClassificationDataLoader, metrics.FlopsWeightedAccuracyMetric
    ):
        def __init__(self, hparams):
            self.lin = torch.nn.Linear(self.input_shape[0], self.output_shape)
            self.loss = torch.nn.functional.mse_loss

        def forward(self, x):
            x = x.view(x.size(0), -1)
            x = torch.relu(self.lin(x))
            return x

    net = DummyModule()
    targets = torch.tensor([[0], [0], [0], [0], [1], [1], [1], [1]])
    preds = torch.tensor([[0.1, 0.9] for _ in range(8)])

    assert DummyModule.metric_names() == ["flops", "flops_weighted_acc", "loss"]
    assert net.metrics_step(preds, targets)["flops"] == torch.tensor(22.0)
    assert torch.tensor(0.0) < net.metrics_step(preds, targets)["flops_weighted_acc"]


def test_MetricSelector():
    class DummyModule(
        RideModule,
        DummyClassificationDataLoader,
        metrics.MetricSelector(
            {
                "one": metrics.TopKAccuracyMetric(1),
                "two": [
                    metrics.FlopsWeightedAccuracyMetric,
                    metrics.TopKAccuracyMetric(2),
                ],
            }
        ),
    ):
        def __init__(self, hparams):
            self.lin = torch.nn.Linear(self.input_shape[0], self.output_shape)
            self.loss = torch.nn.functional.mse_loss

        def forward(self, x):
            x = x.view(x.size(0), -1)
            x = torch.relu(self.lin(x))
            return x

    hparams = attributedict(DummyModule.configs().default_values())
    hparams.metric_selection = "two"
    net = DummyModule(hparams)
    targets = torch.tensor([[0], [0], [0], [0], [1], [1], [1], [1]])
    preds = torch.tensor([[0.1, 0.9] for _ in range(8)])

    assert set(DummyModule.metric_names()) == {
        "flops",
        "flops_weighted_acc",
        "loss",
        "top1acc",
        "top2acc",
    }
    step_output = net.metrics_step(preds, targets)
    assert net.metrics_step(preds, targets)["flops"] == torch.tensor(22.0)
    assert torch.tensor(0.0) < step_output["flops_weighted_acc"]
    assert torch.tensor(0.0) < step_output["top2acc"]
    assert "top1acc" not in step_output



import sys

from setuptools import find_packages, setup

try:
    from ride import info
except ImportError:
    # alternative https://stackoverflow.com/a/67692/4521646
    sys.path.append("ride")
    import info  # noqa: F401


def from_file(file_name: str = "requirements.txt", comment_char: str = "#"):
    """Load requirements from a file"""
    with open(file_name, "r") as file:
        lines = [ln.strip() for ln in file.readlines()]
    reqs = []
    for ln in lines:
        # filer all comments
        if comment_char in ln:
            ln = ln[: ln.index(comment_char)].strip()
        # skip directly installed dependencies
        if ln.startswith("http"):
            continue
        if ln:  # if requirement is not empty
            reqs.append(ln)
    return reqs


def long_description():
    text = open("README.md", encoding="utf-8").read()
    # SVG images are not readable on PyPI, so replace them  with PNG
    text = text.replace(".svg", ".png")
    return text


# Override name for test version, since `ride` was already taken on TestPyPI
if "--test" in sys.argv:
    sys.argv.pop(sys.argv.index("--test"))
    name = "ride-the-lightning"
else:
    name = "ride"

setup(
    name=name,
    version=info.__version__,
    description=info.__docs__,
    long_description=long_description(),
    long_description_content_type="text/markdown",
    author=info.__author__,
    author_email=info.__author_email__,
    url=info.__homepage__,
    install_requires=from_file("requirements.txt"),
    extras_require={
        "build": from_file("requirements/build.txt"),
        "dev": from_file("requirements/dev.txt"),
        "docs": from_file("requirements/docs.txt"),
        "extras": from_file("requirements/extras.txt"),
    },
    packages=find_packages(exclude=["test"]),
    keywords=["deep learning", "pytorch", "AI"],
    classifiers=[
        "Environment :: Console",
        "Natural Language :: English",
        # How mature is this project? Common values are
        #   3 - Alpha, 4 - Beta, 5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Scientific/Engineering :: Information Analysis",
        # Pick your license as you wish
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



import logging
from pathlib import Path
from typing import Tuple

import pytest
import torch

from ride import Configs, Main
from ride.core import RideModule
from ride.feature_visualisation import FeatureVisualisable
from ride.optimizers import SgdOptimizer
from ride.utils.io import is_nonempty_file
from ride.utils.utils import AttributeDict

from .dummy_dataset import DummyClassificationDataLoader, DummyRegressionDataLoader


class VisDummyModule(
    RideModule, FeatureVisualisable, DummyClassificationDataLoader, SgdOptimizer
):
    def __init__(self, hparams):
        self.conv = torch.nn.Conv2d(3, 5, (3, 3))
        self.lin = torch.nn.Linear(45, self.output_shape)

    def forward(self, x):
        x = torch.relu(self.conv(x.reshape(-1, 3, 5, 5)))
        x = x.view(x.size(0), -1)
        x = torch.relu(self.lin(x))
        return x

    @staticmethod
    def configs():
        c = Configs()
        c.add(
            name="input_shape",
            type=int,
            default=3 * 5 * 5,
            strategy="constant",
            description="Input shape for data.",
        )
        return c


@pytest.fixture()  # scope="module"
def main_and_args() -> Tuple[Main, AttributeDict]:
    m = Main(VisDummyModule)
    parser = m.argparse(run=False)
    args, _ = parser.parse_known_args()
    args.max_epochs = 1
    args.gpus = 0
    args.checkpoint_callback = True
    args.optimization_metric = "loss"
    args.id = "automated_test"
    args.test_ensemble = 0
    args.checkpoint_every_n_steps = 0
    args.monitor_lr = 0
    args.auto_lr_find = 0
    args.auto_scale_batch_size = 0
    args.num_workers = 1
    args.batch_size = 4

    return m, args


class VisDummyModule2(
    RideModule, FeatureVisualisable, DummyRegressionDataLoader, SgdOptimizer
):
    def __init__(self, hparams):
        self.l1 = torch.nn.Linear(self.input_shape[0], self.hparams.hidden_dim)
        self.l2 = torch.nn.Linear(self.hparams.hidden_dim, self.output_shape)
        self.loss = torch.nn.functional.mse_loss

    def forward(self, x):
        x = x.view(x.size(0), -1)
        x = torch.relu(self.l1(x))
        x = torch.relu(self.l2(x))
        return x

    @staticmethod
    def configs():
        c = Configs()
        c.add(
            name="hidden_dim",
            type=int,
            default=128,
            strategy="choice",
            choices=[128, 256, 512, 1024],
            description="Number of hiden units.",
        )
        return c


@pytest.fixture()  # scope="module"
def main_and_args2() -> Tuple[Main, AttributeDict]:
    m = Main(VisDummyModule2)
    parser = m.argparse(run=False)
    args, _ = parser.parse_known_args()
    args.max_epochs = 1
    args.gpus = 0
    args.checkpoint_callback = True
    args.optimization_metric = "loss"
    args.id = "automated_test"
    args.test_ensemble = 0
    args.checkpoint_every_n_steps = 0
    args.monitor_lr = 0
    args.auto_lr_find = 0
    args.auto_scale_batch_size = 0
    args.num_workers = 1
    args.batch_size = 4

    return m, args


class TestFeatureExtraction:
    def test_2d_feat_pca(self, caplog, main_and_args: Tuple[Main, AttributeDict]):
        m, args = main_and_args
        args.test = True
        args.extract_features_after_layer = "conv"
        args.visualise_features = "pca"

        with caplog.at_level(logging.DEBUG):
            m.main(args)

        # Paths are printed in main
        for check in ["conv.npy", "conv_pca.npy", "conv_pca.png"]:
            assert any([check in msg for msg in caplog.messages])

        # Files exist
        d = Path(m.log_dir)
        assert is_nonempty_file(d / "features" / "test" / "conv.npy")
        assert is_nonempty_file(d / "features" / "test" / "conv_pca.npy")
        assert is_nonempty_file(d / "figures" / "test" / "conv_pca.png")

    def test_1d_feat_tse(self, caplog, main_and_args: Tuple[Main, AttributeDict]):
        m, args = main_and_args
        args.test = True
        args.extract_features_after_layer = "lin"
        args.visualise_features = "tsne"

        with caplog.at_level(logging.DEBUG):
            m.main(args)

        # Paths are printed in main
        for check in ["lin.npy", "lin_tsne.npy", "lin_tsne.png"]:
            assert any([check in msg for msg in caplog.messages])

        # Files exist
        d = Path(m.log_dir)
        assert is_nonempty_file(d / "features" / "test" / "lin.npy")
        assert is_nonempty_file(d / "features" / "test" / "lin_tsne.npy")
        assert is_nonempty_file(d / "figures" / "test" / "lin_tsne.png")

    def test_1d_feat_umap(self, caplog, main_and_args2: Tuple[Main, AttributeDict]):
        m, args = main_and_args2
        args.test = True
        args.extract_features_after_layer = "l2"
        args.visualise_features = "umap"

        with caplog.at_level(logging.DEBUG):
            m.main(args)

        # Paths are printed in main
        for check in ["l2.npy", "l2_umap.npy", "l2_umap.png"]:
            assert any([check in msg for msg in caplog.messages])

        # Files exist
        d = Path(m.log_dir)
        assert is_nonempty_file(d / "features" / "test" / "l2.npy")
        assert is_nonempty_file(d / "features" / "test" / "l2_umap.npy")
        assert is_nonempty_file(d / "figures" / "test" / "l2_umap.png")



import os
from pathlib import Path


def latest_file_in(path: Path) -> Path:
    assert path.is_dir()
    paths = list(path.glob("*"))
    if len(paths) == 0:
        raise FileNotFoundError(f"No files found in {str(path)}")

    return Path(max(path.glob("*"), key=os.path.getctime))


def get_latest_checkpoint(log_dir: str) -> Path:
    logs_path = Path(log_dir)
    assert logs_path.is_dir()
    checkpoint_path = latest_file_in(logs_path / "checkpoints")
    if not checkpoint_path:
        raise FileNotFoundError(f"No checkpoints found in {str(logs_path)}")
    assert checkpoint_path.exists()
    return checkpoint_path


def find_checkpoint(path: str) -> str:
    path = Path(path)
    assert path.exists()
    if "ckpt" in path.suffix:
        return str(path)

    latest_file = latest_file_in(path)
    if "ckpt" in latest_file.suffix:
        return str(latest_file)

    checkpoint_dir = path / "checkpoints"
    assert checkpoint_dir.exists()

    latest_file = latest_file_in(checkpoint_dir)
    if "ckpt" in latest_file.suffix:
        return str(latest_file)

    raise RuntimeError(f"Unable to find any checkpoint in {str(path)}")



"""
Modules adding optimizers
"""

from math import ceil
from operator import attrgetter
from typing import Callable

import torch

from ride.core import Configs, OptimizerMixin
from ride.utils.discriminative_lr import discriminative_lr


def discounted_steps_per_epoch(
    base_steps: int, num_gpus: int, accumulate_grad_batches: int
):
    return max(
        1, ceil(base_steps / max(1, num_gpus) / max(1, accumulate_grad_batches or 1))
    )


class SgdOptimizer(OptimizerMixin):
    hparams: ...
    parameters: Callable

    def validate_attributes(self):
        attrgetter("parameters")(self)
        for hparam in SgdOptimizer.configs().names:
            attrgetter(f"hparams.{hparam}")(self)

    @staticmethod
    def configs() -> Configs:
        c = Configs()
        c.add(
            name="learning_rate",
            type=float,
            default=0.1,
            choices=(5e-2, 5e-1),
            strategy="loguniform",
            description="Learning rate.",
        )
        c.add(
            name="weight_decay",
            type=float,
            default=1e-5,
            choices=(1e-6, 1e-3),
            strategy="loguniform",
            description="Weight decay.",
        )
        c.add(
            name="momentum",
            type=float,
            default=0.9,
            choices=(0, 0.999),
            strategy="uniform",
            description="Momentum.",
        )
        return c

    def configure_optimizers(self):
        optimizer = torch.optim.SGD(
            self.parameters(),
            lr=self.hparams.learning_rate,
            momentum=self.hparams.momentum,
            weight_decay=self.hparams.weight_decay,
        )
        return optimizer


class AdamWOptimizer(OptimizerMixin):
    hparams: ...
    parameters: Callable

    def validate_attributes(self):
        attrgetter("parameters")(self)
        for hparam in AdamWOptimizer.configs().names:
            attrgetter(f"hparams.{hparam}")(self)

    @staticmethod
    def configs() -> Configs:
        c = Configs()
        c.add(
            name="learning_rate",
            type=float,
            default=0.001,
            choices=(5e-7, 5e-1),
            strategy="loguniform",
            description="Learning rate.",
        )
        c.add(
            name="optimizer_beta1",
            type=float,
            default=0.9,
            choices=(0, 0.999),
            strategy="uniform",
            description="Beta 1.",
        )
        c.add(
            name="optimizer_beta2",
            type=float,
            default=0.999,
            choices=(0, 0.99999),
            strategy="uniform",
            description="Beta 2.",
        )
        c.add(
            name="weight_decay",
            type=float,
            default=1e-2,
            choices=(1e-6, 1e-1),
            strategy="loguniform",
            description="Weight decay.",
        )
        return c

    def configure_optimizers(self):
        optimizer = torch.optim.AdamW(
            self.parameters(),
            lr=self.hparams.learning_rate,
            betas=(self.hparams.optimizer_beta1, self.hparams.optimizer_beta2),
            weight_decay=self.hparams.weight_decay,
        )
        return optimizer


class SgdReduceLrOnPlateauOptimizer(OptimizerMixin):
    hparams: ...
    parameters: Callable

    def validate_attributes(self):
        attrgetter("parameters")(self)
        for hparam in SgdReduceLrOnPlateauOptimizer.configs().names:
            attrgetter(f"hparams.{hparam}")(self)

    @staticmethod
    def configs() -> Configs:
        c = SgdOptimizer.configs()
        c.add(
            name="learning_rate_reduction_patience",
            type=int,
            default=10,
            strategy="choice",
            description="Number of epochs before learning rate should be reduced",
        )
        c.add(
            name="learning_rate_reduction_factor",
            type=float,
            default=0.1,
            strategy="choice",
            description="Reduction factor when learning rate is reduced",
        )
        return c

    def configure_optimizers(self):
        optimizer = torch.optim.SGD(
            self.parameters(),
            lr=self.hparams.learning_rate,
            momentum=self.hparams.momentum,
            weight_decay=self.hparams.weight_decay,
        )

        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            optimizer,
            "min",
            patience=self.hparams.learning_rate_reduction_patience,
            factor=self.hparams.learning_rate_reduction_factor,
            verbose=True,
        )
        return {
            "optimizer": optimizer,
            "lr_scheduler": scheduler,
            "monitor": "train/loss",
        }


class AdamWReduceLrOnPlateauOptimizer(OptimizerMixin):
    hparams: ...
    parameters: Callable

    def validate_attributes(self):
        attrgetter("parameters")(self)
        for hparam in AdamWReduceLrOnPlateauOptimizer.configs().names:
            attrgetter(f"hparams.{hparam}")(self)

    @staticmethod
    def configs() -> Configs:
        c = AdamWOptimizer.configs()
        c.add(
            name="learning_rate_reduction_patience",
            type=int,
            default=10,
            strategy="choice",
            description="Number of epochs before learning rate should be reduced",
        )
        c.add(
            name="learning_rate_reduction_factor",
            type=float,
            default=0.1,
            strategy="choice",
            description="Reduction factor when learning rate is reduced",
        )
        return c

    def configure_optimizers(self):
        optimizer = torch.optim.AdamW(
            self.parameters(),
            lr=self.hparams.learning_rate,
            betas=(self.hparams.optimizer_beta1, self.hparams.optimizer_beta2),
            weight_decay=self.hparams.weight_decay,
        )
        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            optimizer,
            "min",
            patience=self.hparams.learning_rate_reduction_patience,
            factor=self.hparams.learning_rate_reduction_factor,
            verbose=True,
        )
        return {
            "optimizer": optimizer,
            "lr_scheduler": scheduler,
            "monitor": "train/loss",
        }


class SgdCyclicLrOptimizer(OptimizerMixin):
    hparams: ...
    parameters: Callable
    train_dataloader: Callable

    def validate_attributes(self):
        attrgetter("parameters")(self)
        attrgetter("train_dataloader")(self)
        attrgetter("hparams.batch_size")(self)
        attrgetter("hparams.num_gpus")(self)
        attrgetter("hparams.max_epochs")(self)
        attrgetter("hparams.accumulate_grad_batches")(self)
        for hparam in SgdCyclicLrOptimizer.configs().names:
            attrgetter(f"hparams.{hparam}")(self)

    @staticmethod
    def configs() -> Configs:
        c = SgdOptimizer.configs()
        c.add(
            name="discriminative_lr_fraction",
            type=float,
            default=1,
            choices=(1e-7, 1),
            strategy="loguniform",
            description=(
                "Discriminative learning rate fraction of early layers compared to final layers. "
                "If `1`, discriminative learning rate is not used."
            ),
        )
        return c

    def configure_optimizers(self):
        params, lr = discriminative_lr_and_params(
            self, self.hparams.learning_rate, self.hparams.discriminative_lr_fraction
        )
        optimizer = torch.optim.SGD(
            params=params,
            lr=self.hparams.learning_rate,
            momentum=self.hparams.momentum,
            weight_decay=self.hparams.weight_decay,
        )
        # Use recommendations from https://arxiv.org/abs/1506.01186
        base_lr = [x / 4 for x in lr] if isinstance(lr, list) else lr / 4
        scheduler = torch.optim.lr_scheduler.CyclicLR(
            optimizer,
            base_lr=base_lr,
            max_lr=lr,
            step_size_up=discounted_steps_per_epoch(
                len(self.train_dataloader()) / 4,
                self.hparams.num_gpus,
                self.hparams.accumulate_grad_batches,
            ),
            step_size_down=discounted_steps_per_epoch(
                (len(self.train_dataloader()) - len(self.train_dataloader()) / 4),
                self.hparams.num_gpus,
                self.hparams.accumulate_grad_batches,
            ),
            cycle_momentum=True,  # Not supported
        )
        return [optimizer], {"scheduler": scheduler, "interval": "step"}


class AdamWCyclicLrOptimizer(OptimizerMixin):
    hparams: ...
    parameters: Callable
    train_dataloader: Callable

    def validate_attributes(self):
        attrgetter("parameters")(self)
        attrgetter("train_dataloader")(self)
        attrgetter("hparams.batch_size")(self)
        attrgetter("hparams.max_epochs")(self)
        attrgetter("hparams.accumulate_grad_batches")(self)
        for hparam in AdamWCyclicLrOptimizer.configs().names:
            attrgetter(f"hparams.{hparam}")(self)

    @staticmethod
    def configs() -> Configs:
        c = AdamWOptimizer.configs()
        c.add(
            name="discriminative_lr_fraction",
            type=float,
            default=1,
            choices=(1e-7, 1),
            strategy="loguniform",
            description=(
                "Discriminative learning rate fraction of early layers compared to final layers. "
                "If `1`, discriminative learning rate is not used."
            ),
        )
        return c

    def configure_optimizers(self):
        params, lr = discriminative_lr_and_params(
            self, self.hparams.learning_rate, self.hparams.discriminative_lr_fraction
        )
        optimizer = torch.optim.AdamW(
            params,
            lr=self.hparams.learning_rate,
            betas=(self.hparams.optimizer_beta1, self.hparams.optimizer_beta2),
            weight_decay=self.hparams.weight_decay,
        )
        # Use recommendations from https://arxiv.org/abs/1506.01186
        base_lr = [x / 4 for x in lr] if isinstance(lr, list) else lr / 4
        scheduler = torch.optim.lr_scheduler.CyclicLR(
            optimizer,
            base_lr=base_lr,
            max_lr=lr,
            step_size_up=discounted_steps_per_epoch(
                len(self.train_dataloader()) / 4,
                self.hparams.num_gpus,
                self.hparams.accumulate_grad_batches,
            ),
            step_size_down=discounted_steps_per_epoch(
                len(self.train_dataloader()) - len(self.train_dataloader()) / 4,
                self.hparams.num_gpus,
                self.hparams.accumulate_grad_batches,
            ),
            cycle_momentum=False,
        )
        return [optimizer], {"scheduler": scheduler, "interval": "step"}


class SgdOneCycleOptimizer(OptimizerMixin):
    hparams: ...
    parameters: Callable
    train_dataloader: Callable

    def validate_attributes(self):
        attrgetter("parameters")(self)
        attrgetter("train_dataloader")(self)
        attrgetter("hparams.max_epochs")(self)
        attrgetter("hparams.batch_size")(self)
        attrgetter("hparams.num_gpus")(self)
        attrgetter("hparams.accumulate_grad_batches")(self)
        for hparam in SgdOneCycleOptimizer.configs().names:
            attrgetter(f"hparams.{hparam}")(self)

    @staticmethod
    def configs() -> Configs:
        c = SgdOptimizer.configs()
        c.add(
            name="discriminative_lr_fraction",
            type=float,
            default=1,
            choices=(1e-7, 1),
            strategy="loguniform",
            description=(
                "Discriminative learning rate fraction of early layers compared to final layers. "
                "If `1`, discriminative learning rate is not used."
            ),
        )
        return c

    def configure_optimizers(self):
        params, lr = discriminative_lr_and_params(
            self, self.hparams.learning_rate, self.hparams.discriminative_lr_fraction
        )
        optimizer = torch.optim.SGD(
            params=params,
            lr=self.hparams.learning_rate,
            momentum=self.hparams.momentum,
            weight_decay=self.hparams.weight_decay,
        )
        scheduler = torch.optim.lr_scheduler.OneCycleLR(
            optimizer,
            max_lr=lr,
            steps_per_epoch=discounted_steps_per_epoch(
                len(self.train_dataloader()),
                self.hparams.num_gpus,
                self.hparams.accumulate_grad_batches,
            ),
            epochs=self.hparams.max_epochs,
        )
        return [optimizer], {"scheduler": scheduler, "interval": "step"}


class AdamWOneCycleOptimizer(OptimizerMixin):
    hparams: ...
    parameters: Callable
    train_dataloader: Callable

    def validate_attributes(self):
        attrgetter("parameters")(self)
        attrgetter("train_dataloader")(self)
        attrgetter("hparams.batch_size")(self)
        attrgetter("hparams.max_epochs")(self)
        attrgetter("hparams.num_gpus")(self)
        attrgetter("hparams.accumulate_grad_batches")(self)
        for hparam in AdamWOneCycleOptimizer.configs().names:
            attrgetter(f"hparams.{hparam}")(self)

    @staticmethod
    def configs() -> Configs:
        c = AdamWOptimizer.configs()
        c.add(
            name="discriminative_lr_fraction",
            type=float,
            default=1,
            choices=(1e-7, 1),
            strategy="loguniform",
            description=(
                "Discriminative learning rate fraction of early layers compared to final layers. "
                "If `1`, discriminative learning rate is not used."
            ),
        )
        return c

    def configure_optimizers(self):
        params, lr = discriminative_lr_and_params(
            self, self.hparams.learning_rate, self.hparams.discriminative_lr_fraction
        )
        optimizer = torch.optim.AdamW(
            params=params,
            lr=self.hparams.learning_rate,
            betas=(self.hparams.optimizer_beta1, self.hparams.optimizer_beta2),
            weight_decay=self.hparams.weight_decay,
        )
        scheduler = torch.optim.lr_scheduler.OneCycleLR(
            optimizer,
            max_lr=lr,
            steps_per_epoch=discounted_steps_per_epoch(
                len(self.train_dataloader()),
                self.hparams.num_gpus,
                self.hparams.accumulate_grad_batches,
            ),
            epochs=self.hparams.max_epochs,
        )
        return [optimizer], {"scheduler": scheduler, "interval": "step"}


def discriminative_lr_and_params(
    model: torch.nn.Module, lr: float, discriminative_lr_fraction: float
):
    if discriminative_lr_fraction != 1:
        params, max_lr, _ = discriminative_lr(
            model,
            slice(
                lr * discriminative_lr_fraction,
                lr,
            ),
        )
    else:
        params = model.parameters()
        max_lr = lr

    return params, max_lr



import json
import shutil
from pathlib import Path

from ride.utils import io


def test_bump_version():
    temp_dir = Path(__file__).parent / "temp"
    temp_dir.mkdir(exist_ok=True, parents=True)

    def write_data(path: Path):
        with path.open("w", encoding="utf-8") as f:
            json.dump({"dummy": 42}, f)

    p1 = temp_dir / "dummy_file.json"
    p2 = temp_dir / "dummy_file_1.json"
    p3 = temp_dir / "dummy_file_2.json"

    # Ensure files don't exist beforehand
    try:
        p1.unlink()
        p2.unlink()
        p3.unlink()
    except FileNotFoundError:
        pass

    # Non existing file
    assert io.bump_version(p1) == p1
    write_data(p1)
    assert p1.exists()

    # File exists (with no '_x' prepended)
    assert not p2.exists()
    assert io.bump_version(p1) == p2
    write_data(p2)
    assert p2.exists()

    # File exists (with '_x' prepended)
    assert not p3.exists()
    assert io.bump_version(p1) == p3
    assert io.bump_version(p2) == p3

    # Clean up
    shutil.rmtree(str(temp_dir))



import os
from platform import uname
from typing import Any, Dict

from cpuinfo import get_cpu_info
from GPUtil import getGPUs
from psutil import cpu_count, cpu_freq, virtual_memory

NUM_CPU = os.cpu_count() or 1


def get_machine_info() -> Dict[str, Any]:
    sys = uname()
    cpu = get_cpu_info()
    svmem = virtual_memory()
    gpus = getGPUs()

    return {
        "system": {"system": sys.system, "node": sys.node, "release": sys.release},
        "cpu": {
            "model": cpu["brand_raw"],
            "architecture": cpu["arch_string_raw"],
            "cores": {
                "physical": cpu_count(logical=False),
                "total": cpu_count(logical=True),
            },
            "frequency": f"{(cpu_freq().max / 1000):.2f} GHz",
        },
        "memory": {
            "total": get_size(svmem.total),
            "used": get_size(svmem.used),
            "available": get_size(svmem.available),
        },
        "gpus": (
            [{"name": g.name, "memory": f"{g.memoryTotal} MB"} for g in gpus]
            if gpus
            else None
        ),
    }


def get_size(bytes, suffix="B"):
    """Scale bytes to its proper format, e.g. 1253656 => '1.20MB'"""
    factor = 1024
    for unit in ["", "K", "M", "G", "T", "P"]:
        if bytes < factor:
            return f"{bytes:.2f} {unit}{suffix}"
        bytes /= factor
